# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value import ArrayVellumValue
import typing
from .code_execution_node_result_data import CodeExecutionNodeResultData
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class CodeExecutionNodeResult(UniversalBaseModel):
    """
    A Node Result Event emitted from a Code Execution Node.
    """

    type: typing.Literal["CODE_EXECUTION"] = "CODE_EXECUTION"
    data: CodeExecutionNodeResultData

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from ...core.request_options import RequestOptions
from ...core.http_response import HttpResponse
from ...types.deployment_read import DeploymentRead
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from json.decoder import JSONDecodeError
from ...core.api_error import ApiError
from ...types.named_scenario_input_request import NamedScenarioInputRequest
from ...types.sandbox_scenario import SandboxScenario
from ...core.serialization import convert_and_respect_annotation_metadata
from ...core.client_wrapper import AsyncClientWrapper
from ...core.http_response import AsyncHttpResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawSandboxesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def deploy_prompt(
        self,
        id: str,
        prompt_variant_id: str,
        *,
        prompt_deployment_id: typing.Optional[str] = OMIT,
        prompt_deployment_name: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        release_tags: typing.Optional[typing.Sequence[str]] = OMIT,
        release_description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[DeploymentRead]:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this sandbox.

        prompt_variant_id : str
            An ID identifying the Prompt you'd like to deploy.

        prompt_deployment_id : typing.Optional[str]
            The Vellum-generated ID of the Prompt Deployment you'd like to update. Cannot specify both this and prompt_deployment_name. Leave null to create a new Prompt Deployment.

        prompt_deployment_name : typing.Optional[str]
            The unique name of the Prompt Deployment you'd like to either create or update. Cannot specify both this and prompt_deployment_id. If provided and matches an existing Prompt Deployment, that Prompt Deployment will be updated. Otherwise, a new Prompt Deployment will be created.

        label : typing.Optional[str]
            In the event that a new Prompt Deployment is created, this will be the label it's given.

        release_tags : typing.Optional[typing.Sequence[str]]
            Optionally provide the release tags that you'd like to be associated with the latest release of the created/updated Prompt Deployment.

        release_description : typing.Optional[str]
            Optionally provide a description that details what's new in this Release.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeploymentRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/sandboxes/{jsonable_encoder(id)}/prompts/{jsonable_encoder(prompt_variant_id)}/deploy",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "prompt_deployment_id": prompt_deployment_id,
                "prompt_deployment_name": prompt_deployment_name,
                "label": label,
                "release_tags": release_tags,
                "release_description": release_description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeploymentRead,
                    parse_obj_as(
                        type_=DeploymentRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def upsert_sandbox_scenario(
        self,
        id: str,
        *,
        inputs: typing.Sequence[NamedScenarioInputRequest],
        label: typing.Optional[str] = OMIT,
        scenario_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[SandboxScenario]:
        """
        Upserts a new scenario for a sandbox, keying off of the optionally provided scenario id.

        If an id is provided and has a match, the scenario will be updated. If no id is provided or no match
        is found, a new scenario will be appended to the end.

        Note that a full replacement of the scenario is performed, so any fields not provided will be removed
        or overwritten with default values.

        Parameters
        ----------
        id : str
            A UUID string identifying this sandbox.

        inputs : typing.Sequence[NamedScenarioInputRequest]
            The inputs for the scenario

        label : typing.Optional[str]

        scenario_id : typing.Optional[str]
            The id of the scenario to update. If none is provided, an id will be generated and a new scenario will be appended.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[SandboxScenario]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/sandboxes/{jsonable_encoder(id)}/scenarios",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "label": label,
                "inputs": convert_and_respect_annotation_metadata(
                    object_=inputs, annotation=typing.Sequence[NamedScenarioInputRequest], direction="write"
                ),
                "scenario_id": scenario_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SandboxScenario,
                    parse_obj_as(
                        type_=SandboxScenario,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def delete_sandbox_scenario(
        self, id: str, scenario_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Deletes an existing scenario from a sandbox, keying off of the provided scenario id.

        Parameters
        ----------
        id : str
            A UUID string identifying this sandbox.

        scenario_id : str
            An id identifying the scenario that you'd like to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/sandboxes/{jsonable_encoder(id)}/scenarios/{jsonable_encoder(scenario_id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncRawSandboxesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def deploy_prompt(
        self,
        id: str,
        prompt_variant_id: str,
        *,
        prompt_deployment_id: typing.Optional[str] = OMIT,
        prompt_deployment_name: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        release_tags: typing.Optional[typing.Sequence[str]] = OMIT,
        release_description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[DeploymentRead]:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this sandbox.

        prompt_variant_id : str
            An ID identifying the Prompt you'd like to deploy.

        prompt_deployment_id : typing.Optional[str]
            The Vellum-generated ID of the Prompt Deployment you'd like to update. Cannot specify both this and prompt_deployment_name. Leave null to create a new Prompt Deployment.

        prompt_deployment_name : typing.Optional[str]
            The unique name of the Prompt Deployment you'd like to either create or update. Cannot specify both this and prompt_deployment_id. If provided and matches an existing Prompt Deployment, that Prompt Deployment will be updated. Otherwise, a new Prompt Deployment will be created.

        label : typing.Optional[str]
            In the event that a new Prompt Deployment is created, this will be the label it's given.

        release_tags : typing.Optional[typing.Sequence[str]]
            Optionally provide the release tags that you'd like to be associated with the latest release of the created/updated Prompt Deployment.

        release_description : typing.Optional[str]
            Optionally provide a description that details what's new in this Release.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeploymentRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/sandboxes/{jsonable_encoder(id)}/prompts/{jsonable_encoder(prompt_variant_id)}/deploy",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "prompt_deployment_id": prompt_deployment_id,
                "prompt_deployment_name": prompt_deployment_name,
                "label": label,
                "release_tags": release_tags,
                "release_description": release_description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeploymentRead,
                    parse_obj_as(
                        type_=DeploymentRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def upsert_sandbox_scenario(
        self,
        id: str,
        *,
        inputs: typing.Sequence[NamedScenarioInputRequest],
        label: typing.Optional[str] = OMIT,
        scenario_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[SandboxScenario]:
        """
        Upserts a new scenario for a sandbox, keying off of the optionally provided scenario id.

        If an id is provided and has a match, the scenario will be updated. If no id is provided or no match
        is found, a new scenario will be appended to the end.

        Note that a full replacement of the scenario is performed, so any fields not provided will be removed
        or overwritten with default values.

        Parameters
        ----------
        id : str
            A UUID string identifying this sandbox.

        inputs : typing.Sequence[NamedScenarioInputRequest]
            The inputs for the scenario

        label : typing.Optional[str]

        scenario_id : typing.Optional[str]
            The id of the scenario to update. If none is provided, an id will be generated and a new scenario will be appended.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[SandboxScenario]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/sandboxes/{jsonable_encoder(id)}/scenarios",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "label": label,
                "inputs": convert_and_respect_annotation_metadata(
                    object_=inputs, annotation=typing.Sequence[NamedScenarioInputRequest], direction="write"
                ),
                "scenario_id": scenario_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SandboxScenario,
                    parse_obj_as(
                        type_=SandboxScenario,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def delete_sandbox_scenario(
        self, id: str, scenario_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Deletes an existing scenario from a sandbox, keying off of the provided scenario id.

        Parameters
        ----------
        id : str
            A UUID string identifying this sandbox.

        scenario_id : str
            An id identifying the scenario that you'd like to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/sandboxes/{jsonable_encoder(id)}/scenarios/{jsonable_encoder(scenario_id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



import logging
import time
from typing import Any, Dict, Optional

import httpx

from vellum.workflows.emitters.base import BaseWorkflowEmitter
from vellum.workflows.events.types import default_serializer
from vellum.workflows.events.workflow import WorkflowEvent
from vellum.workflows.state.base import BaseState

logger = logging.getLogger(__name__)


class VellumEmitter(BaseWorkflowEmitter):
    """
    Emitter that sends workflow events to Vellum's infrastructure for monitoring
    externally hosted SDK-powered workflows.

    Usage:
        class MyWorkflow(BaseWorkflow):
            emitters = [VellumEmitter]

    The emitter will automatically use the same Vellum client configuration
    as the workflow it's attached to.
    """

    def __init__(
        self,
        *,
        timeout: Optional[float] = 30.0,
        max_retries: int = 3,
    ):
        """
        Initialize the VellumEmitter.

        Args:
            timeout: Request timeout in seconds.
            max_retries: Maximum number of retry attempts for failed requests.
        """
        super().__init__()
        self._timeout = timeout
        self._max_retries = max_retries
        self._events_endpoint = "v1/events"  # TODO: make this configurable with the correct url

    def emit_event(self, event: WorkflowEvent) -> None:
        """
        Emit a workflow event to Vellum's infrastructure.

        Args:
            event: The workflow event to emit.
        """
        if not self._context:
            return

        try:
            event_data = default_serializer(event)

            self._send_event(event_data)

        except Exception as e:
            logger.exception(f"Failed to emit event {event.name}: {e}")

    def snapshot_state(self, state: BaseState) -> None:
        """
        Send a state snapshot to Vellum's infrastructure.

        Args:
            state: The workflow state to snapshot.
        """
        pass

    def _send_event(self, event_data: Dict[str, Any]) -> None:
        """
        Send event data to Vellum's events endpoint with retry logic.

        Args:
            event_data: The serialized event data to send.
        """
        if not self._context:
            logger.warning("Cannot send event: No workflow context registered")
            return

        client = self._context.vellum_client

        for attempt in range(self._max_retries + 1):
            try:
                # Use the Vellum client's underlying HTTP client to make the request
                # For proper authentication headers and configuration
                base_url = client._client_wrapper.get_environment().default
                response = client._client_wrapper.httpx_client.request(
                    method="POST",
                    base_url=base_url,
                    path=self._events_endpoint,  # TODO: will be replaced with the correct url
                    json=event_data,
                    headers=client._client_wrapper.get_headers(),
                    request_options={"timeout_in_seconds": self._timeout},
                )

                response.raise_for_status()

                if attempt > 0:
                    logger.info(f"Event sent successfully after {attempt + 1} attempts")
                return

            except httpx.HTTPStatusError as e:
                if e.response.status_code >= 500:
                    # Server errors might be transient, retry
                    if attempt < self._max_retries:
                        wait_time = min(2**attempt, 60)  # Exponential backoff, max 60s
                        logger.warning(
                            f"Server error emitting event (attempt {attempt + 1}/{self._max_retries + 1}): "
                            f"{e.response.status_code}. Retrying in {wait_time}s..."
                        )
                        time.sleep(wait_time)
                        continue
                    else:
                        logger.exception(
                            f"Server error emitting event after {self._max_retries + 1} attempts: "
                            f"{e.response.status_code} {e.response.text}"
                        )
                        return
                else:
                    # Client errors (4xx) are not retriable
                    logger.exception(f"Client error emitting event: {e.response.status_code} {e.response.text}")
                    return

            except httpx.RequestError as e:
                if attempt < self._max_retries:
                    wait_time = min(2**attempt, 60)  # Exponential backoff, max 60s
                    logger.warning(
                        f"Network error emitting event (attempt {attempt + 1}/{self._max_retries + 1}): "
                        f"{e}. Retrying in {wait_time}s..."
                    )
                    time.sleep(wait_time)
                    continue
                else:
                    logger.exception(f"Network error emitting event after {self._max_retries + 1} attempts: {e}")
                    return



from typing import Any, Dict, cast

from vellum.workflows import BaseWorkflow
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.core.retry_node.node import RetryNode
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_retry_node_parameters():
    """Test that RetryNode parameters are correctly serialized."""

    # GIVEN a RetryNode with specific parameters
    @RetryNode.wrap(max_attempts=5, delay=2.5, retry_on_error_code=WorkflowErrorCode.INVALID_INPUTS)
    class MyRetryNode(BaseNode):
        pass

    # AND a workflow using the node
    class MyWorkflow(BaseWorkflow):
        graph = MyRetryNode

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    serialized_workflow = cast(Dict[str, Any], workflow_display.serialize())

    # THEN the correct inputs should be serialized on the node
    serialized_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] != "ENTRYPOINT"
    )

    retry_adornment = next(
        adornment for adornment in serialized_node["adornments"] if adornment["label"] == "RetryNode"
    )

    max_attempts_attribute = next(attr for attr in retry_adornment["attributes"] if attr["name"] == "max_attempts")
    assert max_attempts_attribute["value"]["value"]["value"] == 5

    delay_attribute = next(attr for attr in retry_adornment["attributes"] if attr["name"] == "delay")
    assert delay_attribute["value"]["value"]["value"] == 2.5

    retry_on_error_code_attribute = next(
        attr for attr in retry_adornment["attributes"] if attr["name"] == "retry_on_error_code"
    )

    assert retry_on_error_code_attribute["value"]["value"]["type"] == "STRING"
    assert retry_on_error_code_attribute["value"]["value"]["value"] == "INVALID_INPUTS"



import pytest

from vellum.workflows.expressions.greater_than import GreaterThanExpression
from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression
from vellum.workflows.expressions.less_than import LessThanExpression
from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression
from vellum.workflows.state.base import BaseState


class Comparable:
    """A custom class with two values, where comparisons use a computed metric (multiplication)."""

    def __init__(self, value1, value2):
        self.value1 = value1  # First numerical value
        self.value2 = value2  # Second numerical value

    def computed_value(self):
        return self.value1 * self.value2  # Multiply for comparison

    def __ge__(self, other):
        if isinstance(other, Comparable):
            return self.computed_value() >= other.computed_value()
        elif isinstance(other, (int, float)):
            return self.computed_value() >= other
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Comparable):
            return self.computed_value() > other.computed_value()
        elif isinstance(other, (int, float)):
            return self.computed_value() > other
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, Comparable):
            return self.computed_value() <= other.computed_value()
        elif isinstance(other, (int, float)):
            return self.computed_value() <= other
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Comparable):
            return self.computed_value() < other.computed_value()
        elif isinstance(other, (int, float)):
            return self.computed_value() < other
        return NotImplemented


class NonComparable:
    """A custom class that does not support comparisons."""

    def __init__(self, value1, value2):
        self.value1 = value1
        self.value2 = value2


class TestState(BaseState):
    pass


def test_greater_than_or_equal_to():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)  # Computed: 4 × 5 = 20
    obj2 = Comparable(2, 10)  # Computed: 2 × 10 = 20
    obj3 = Comparable(3, 6)  # Computed: 3 × 6 = 18
    obj4 = Comparable(5, 5)  # Computed: 5 × 5 = 25

    state = TestState()

    # WHEN comparing objects
    assert GreaterThanOrEqualToExpression(lhs=obj1, rhs=obj2).resolve(state) is True  # 20 >= 20
    assert GreaterThanOrEqualToExpression(lhs=obj1, rhs=obj3).resolve(state) is True  # 20 >= 18
    assert GreaterThanOrEqualToExpression(lhs=obj3, rhs=obj4).resolve(state) is False  # 18 < 25

    # WHEN comparing to raw numbers
    assert GreaterThanOrEqualToExpression(lhs=obj1, rhs=19).resolve(state) is True  # 20 >= 19
    assert GreaterThanOrEqualToExpression(lhs=obj3, rhs=20).resolve(state) is False  # 18 < 20


def test_greater_than_or_equal_to_invalid():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = "invalid"

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanOrEqualToExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>=' not supported between instances of 'Comparable' and 'str'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanOrEqualToExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>=' not supported between instances of 'str' and 'Comparable'"


def test_greater_than_or_equal_to_non_comparable():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = NonComparable(2, 10)

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanOrEqualToExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>=' not supported between instances of 'Comparable' and 'NonComparable'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanOrEqualToExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>=' not supported between instances of 'NonComparable' and 'Comparable'"


def test_greater_than():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)  # Computed: 4 × 5 = 20
    obj2 = Comparable(2, 10)  # Computed: 2 × 10 = 20
    obj3 = Comparable(3, 6)  # Computed: 3 × 6 = 18
    obj4 = Comparable(5, 5)  # Computed: 5 × 5 = 25

    state = TestState()

    # WHEN comparing objects
    assert GreaterThanExpression(lhs=obj1, rhs=obj2).resolve(state) is False  # 20 > 20
    assert GreaterThanExpression(lhs=obj1, rhs=obj3).resolve(state) is True  # 20 > 18
    assert GreaterThanExpression(lhs=obj3, rhs=obj4).resolve(state) is False  # 18 < 25

    # WHEN comparing to raw numbers
    assert GreaterThanExpression(lhs=obj1, rhs=19).resolve(state) is True  # 20 > 19
    assert GreaterThanExpression(lhs=obj3, rhs=20).resolve(state) is False  # 18 < 20


def test_greater_than_invalid():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = "invalid"

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>' not supported between instances of 'Comparable' and 'str'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>' not supported between instances of 'str' and 'Comparable'"


def test_greater_than_non_comparable():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = NonComparable(2, 10)

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>' not supported between instances of 'Comparable' and 'NonComparable'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        GreaterThanExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'>' not supported between instances of 'NonComparable' and 'Comparable'"


def test_less_than_or_equal_to():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)  # Computed: 4 × 5 = 20
    obj2 = Comparable(2, 10)  # Computed: 2 × 10 = 20
    obj3 = Comparable(3, 6)  # Computed: 3 × 6 = 18
    obj4 = Comparable(5, 5)  # Computed: 5 × 5 = 25

    state = TestState()

    # WHEN comparing objects
    assert LessThanOrEqualToExpression(lhs=obj1, rhs=obj2).resolve(state) is True  # 20 <= 20
    assert LessThanOrEqualToExpression(lhs=obj1, rhs=obj3).resolve(state) is False  # 20 > 18
    assert LessThanOrEqualToExpression(lhs=obj3, rhs=obj4).resolve(state) is True  # 18 <= 25

    # WHEN comparing to raw numbers
    assert LessThanOrEqualToExpression(lhs=obj1, rhs=21).resolve(state) is True  # 20 <= 21
    assert LessThanOrEqualToExpression(lhs=obj3, rhs=17).resolve(state) is False  # 18 > 17


def test_less_than_or_equal_to_invalid():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = "invalid"

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanOrEqualToExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<=' not supported between instances of 'Comparable' and 'str'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanOrEqualToExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<=' not supported between instances of 'str' and 'Comparable'"


def test_less_than_or_equal_to_non_comparable():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = NonComparable(2, 10)

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanOrEqualToExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<=' not supported between instances of 'Comparable' and 'NonComparable'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanOrEqualToExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<=' not supported between instances of 'NonComparable' and 'Comparable'"


def test_less_than():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)  # Computed: 4 × 5 = 20
    obj2 = Comparable(2, 10)  # Computed: 2 × 10 = 20
    obj3 = Comparable(3, 6)  # Computed: 3 × 6 = 18
    obj4 = Comparable(5, 5)  # Computed: 5 × 5 = 25

    state = TestState()

    # WHEN comparing objects
    assert LessThanExpression(lhs=obj1, rhs=obj2).resolve(state) is False  # 20 < 20
    assert LessThanExpression(lhs=obj1, rhs=obj3).resolve(state) is False  # 20 < 18
    assert LessThanExpression(lhs=obj3, rhs=obj4).resolve(state) is True  # 18 < 25

    # WHEN comparing to raw numbers
    assert LessThanExpression(lhs=obj1, rhs=21).resolve(state) is True  # 20 < 21
    assert LessThanExpression(lhs=obj3, rhs=17).resolve(state) is False  # 18 > 17


def test_less_than_invalid():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = "invalid"

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<' not supported between instances of 'Comparable' and 'str'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<' not supported between instances of 'str' and 'Comparable'"


def test_less_than_non_comparable():
    # GIVEN objects with two values
    obj1 = Comparable(4, 5)
    obj2 = NonComparable(2, 10)

    state = TestState()

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanExpression(lhs=obj1, rhs=obj2).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<' not supported between instances of 'Comparable' and 'NonComparable'"

    # WHEN comparing objects with incompatible types
    with pytest.raises(TypeError) as exc_info:
        LessThanExpression(lhs=obj2, rhs=obj1).resolve(state)

    # THEN the expected error is raised
    assert str(exc_info.value) == "'<' not supported between instances of 'NonComparable' and 'Comparable'"



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .folder_entity_test_suite_data import FolderEntityTestSuiteData
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class FolderEntityTestSuite(UniversalBaseModel):
    """
    A slim representation of a Test Suite, as it exists within a Folder.
    """

    id: str
    type: typing.Literal["TEST_SUITE"] = "TEST_SUITE"
    data: FolderEntityTestSuiteData

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from .raw_client import RawDocumentsClient
from ...core.request_options import RequestOptions
from ...types.paginated_slim_document_list import PaginatedSlimDocumentList
from ...types.document_read import DocumentRead
from ...types.document_status import DocumentStatus
from ... import core
from ...types.upload_document_response import UploadDocumentResponse
from ...core.client_wrapper import AsyncClientWrapper
from .raw_client import AsyncRawDocumentsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class DocumentsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawDocumentsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawDocumentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawDocumentsClient
        """
        return self._raw_client

    def list(
        self,
        *,
        document_index_id: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedSlimDocumentList:
        """
        Used to list documents. Optionally filter on supported fields.

        Parameters
        ----------
        document_index_id : typing.Optional[str]
            Filter down to only those documents that are included in the specified index. You may provide either the Vellum-generated ID or the unique name of the index specified upon initial creation.

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        search : typing.Optional[str]
            A search term.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedSlimDocumentList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.documents.list()
        """
        response = self._raw_client.list(
            document_index_id=document_index_id,
            limit=limit,
            offset=offset,
            ordering=ordering,
            search=search,
            request_options=request_options,
        )
        return response.data

    def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> DocumentRead:
        """
        Retrieve a Document, keying off of either its Vellum-generated ID or its external ID.

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.documents.retrieve(
            id="id",
        )
        """
        response = self._raw_client.retrieve(id, request_options=request_options)
        return response.data

    def destroy(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete a Document, keying off of either its Vellum-generated ID or its external ID.

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.documents.destroy(
            id="id",
        )
        """
        response = self._raw_client.destroy(id, request_options=request_options)
        return response.data

    def partial_update(
        self,
        id: str,
        *,
        label: typing.Optional[str] = OMIT,
        status: typing.Optional[DocumentStatus] = OMIT,
        keywords: typing.Optional[typing.Sequence[str]] = OMIT,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentRead:
        """
        Update a Document, keying off of either its Vellum-generated ID or its external ID. Particularly useful for updating its metadata.

        Parameters
        ----------
        id : str

        label : typing.Optional[str]
            A human-readable label for the document. Defaults to the originally uploaded file's file name.

        status : typing.Optional[DocumentStatus]
            The current status of the document

            * `ACTIVE` - Active

        keywords : typing.Optional[typing.Sequence[str]]
            A list of keywords that'll be associated with the document. Used as part of keyword search.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            A JSON object containing any metadata associated with the document that you'd like to filter upon later.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.documents.partial_update(
            id="id",
        )
        """
        response = self._raw_client.partial_update(
            id, label=label, status=status, keywords=keywords, metadata=metadata, request_options=request_options
        )
        return response.data

    def upload(
        self,
        *,
        label: str,
        contents: core.File,
        add_to_index_names: typing.Optional[typing.List[str]] = OMIT,
        external_id: typing.Optional[str] = OMIT,
        keywords: typing.Optional[typing.List[str]] = OMIT,
        metadata: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UploadDocumentResponse:
        """
        Upload a document to be indexed and used for search.

        **Note:** Uses a base url of `https://documents.vellum.ai`.

        Parameters
        ----------
        label : str
            A human-friendly name for this document. Typically the filename.

        contents : core.File
            See core.File for more documentation

        add_to_index_names : typing.Optional[typing.List[str]]
            Optionally include the names of all indexes that you'd like this document to be included in

        external_id : typing.Optional[str]
            Optionally include an external ID for this document. This is useful if you want to re-upload the same document later when its contents change and would like it to be re-indexed.

        keywords : typing.Optional[typing.List[str]]
            Optionally include a list of keywords that'll be associated with this document. Used when performing keyword searches.

        metadata : typing.Optional[str]
            A stringified JSON object containing any metadata associated with the document that you'd like to filter upon later.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UploadDocumentResponse

        """
        response = self._raw_client.upload(
            label=label,
            contents=contents,
            add_to_index_names=add_to_index_names,
            external_id=external_id,
            keywords=keywords,
            metadata=metadata,
            request_options=request_options,
        )
        return response.data


class AsyncDocumentsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawDocumentsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawDocumentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawDocumentsClient
        """
        return self._raw_client

    async def list(
        self,
        *,
        document_index_id: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedSlimDocumentList:
        """
        Used to list documents. Optionally filter on supported fields.

        Parameters
        ----------
        document_index_id : typing.Optional[str]
            Filter down to only those documents that are included in the specified index. You may provide either the Vellum-generated ID or the unique name of the index specified upon initial creation.

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        search : typing.Optional[str]
            A search term.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedSlimDocumentList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.documents.list()


        asyncio.run(main())
        """
        response = await self._raw_client.list(
            document_index_id=document_index_id,
            limit=limit,
            offset=offset,
            ordering=ordering,
            search=search,
            request_options=request_options,
        )
        return response.data

    async def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> DocumentRead:
        """
        Retrieve a Document, keying off of either its Vellum-generated ID or its external ID.

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.documents.retrieve(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.retrieve(id, request_options=request_options)
        return response.data

    async def destroy(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete a Document, keying off of either its Vellum-generated ID or its external ID.

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.documents.destroy(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.destroy(id, request_options=request_options)
        return response.data

    async def partial_update(
        self,
        id: str,
        *,
        label: typing.Optional[str] = OMIT,
        status: typing.Optional[DocumentStatus] = OMIT,
        keywords: typing.Optional[typing.Sequence[str]] = OMIT,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentRead:
        """
        Update a Document, keying off of either its Vellum-generated ID or its external ID. Particularly useful for updating its metadata.

        Parameters
        ----------
        id : str

        label : typing.Optional[str]
            A human-readable label for the document. Defaults to the originally uploaded file's file name.

        status : typing.Optional[DocumentStatus]
            The current status of the document

            * `ACTIVE` - Active

        keywords : typing.Optional[typing.Sequence[str]]
            A list of keywords that'll be associated with the document. Used as part of keyword search.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            A JSON object containing any metadata associated with the document that you'd like to filter upon later.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.documents.partial_update(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.partial_update(
            id, label=label, status=status, keywords=keywords, metadata=metadata, request_options=request_options
        )
        return response.data

    async def upload(
        self,
        *,
        label: str,
        contents: core.File,
        add_to_index_names: typing.Optional[typing.List[str]] = OMIT,
        external_id: typing.Optional[str] = OMIT,
        keywords: typing.Optional[typing.List[str]] = OMIT,
        metadata: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UploadDocumentResponse:
        """
        Upload a document to be indexed and used for search.

        **Note:** Uses a base url of `https://documents.vellum.ai`.

        Parameters
        ----------
        label : str
            A human-friendly name for this document. Typically the filename.

        contents : core.File
            See core.File for more documentation

        add_to_index_names : typing.Optional[typing.List[str]]
            Optionally include the names of all indexes that you'd like this document to be included in

        external_id : typing.Optional[str]
            Optionally include an external ID for this document. This is useful if you want to re-upload the same document later when its contents change and would like it to be re-indexed.

        keywords : typing.Optional[typing.List[str]]
            Optionally include a list of keywords that'll be associated with this document. Used when performing keyword searches.

        metadata : typing.Optional[str]
            A stringified JSON object containing any metadata associated with the document that you'd like to filter upon later.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UploadDocumentResponse

        """
        response = await self._raw_client.upload(
            label=label,
            contents=contents,
            add_to_index_names=add_to_index_names,
            external_id=external_id,
            keywords=keywords,
            metadata=metadata,
            request_options=request_options,
        )
        return response.data



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import pydantic
import typing
from .test_suite_test_case_deleted_bulk_result_data import TestSuiteTestCaseDeletedBulkResultData
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class TestSuiteTestCaseDeletedBulkResult(UniversalBaseModel):
    """
    The result of a bulk operation that deleted a Test Case.
    """

    id: str = pydantic.Field()
    """
    An ID that maps back to one of the initially supplied operations. Can be used to determine the result of a given operation.
    """

    type: typing.Literal["DELETED"] = "DELETED"
    data: TestSuiteTestCaseDeletedBulkResultData

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_default_state.workflow import BasicDefaultStateWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow that has a simple state definition
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicDefaultStateWorkflow)

    serialized_workflow: dict = workflow_display.serialize()
    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "0bbe085c-31d3-4b48-b74d-d501f592da90",
                "key": "example",
                "type": "STRING",
                "default": {"type": "STRING", "value": "hello"},
                "required": False,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its state variables should be what we expect
    state_variables = serialized_workflow["state_variables"]
    assert len(state_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "812ec99b-1859-4361-b795-228628657bac",
                "key": "example",
                "type": "NUMBER",
                "default": {"type": "NUMBER", "value": 5.0},
                "required": False,
                "extensions": {"color": None},
            },
        ],
        state_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "6e7eeaa5-9559-4ae3-8606-e52ead5805a5",
                "key": "example_input",
                "type": "STRING",
            },
            {
                "id": "e3ae0fe3-7590-4eac-b808-45901d82f2ba",
                "key": "example_state",
                "type": "NUMBER",
            },
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 3
    assert len(workflow_raw_data["nodes"]) == 4

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "32684932-7c7c-4b1c-aed2-553de29bf3f7",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "e4136ee4-a51a-4ca3-9a3a-aa96f5de2347",
        },
        "base": None,
        "definition": None,
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    final_output_node_state = next(
        n for n in workflow_raw_data["nodes"] if n["type"] == "TERMINAL" and n["data"]["name"] == "example_state"
    )
    final_output_node_input = next(
        n for n in workflow_raw_data["nodes"] if n["type"] == "TERMINAL" and n["data"]["name"] == "example_input"
    )
    assert not DeepDiff(
        {
            "id": "27fdaa45-b8ce-464d-be50-cf71cc56bc10",
            "type": "TERMINAL",
            "data": {
                "label": "Final Output",
                "name": "example_state",
                "target_handle_id": "e7a09eb2-c9fb-4d57-b436-9cd9384c8960",
                "output_id": "e3ae0fe3-7590-4eac-b808-45901d82f2ba",
                "output_type": "NUMBER",
                "node_input_id": "8de6a408-cf76-4d04-9845-f75211b611be",
            },
            "inputs": [
                {
                    "id": "8de6a408-cf76-4d04-9845-f75211b611be",
                    "key": "node_input",
                    "value": {
                        "rules": [
                            {
                                "type": "NODE_OUTPUT",
                                "data": {
                                    "node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                    "output_id": "84b59a1a-82bf-46bb-9826-9b393402d0fe",
                                },
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "display_data": {"position": {"x": 400.0, "y": -175.0}},
            "base": {
                "name": "FinalOutputNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
            },
            "definition": None,
        },
        final_output_node_state,
    )

    assert not DeepDiff(
        {
            "id": "ca8bb585-c9a8-4bf7-bf9d-534b600fe23b",
            "type": "TERMINAL",
            "data": {
                "label": "Final Output",
                "name": "example_input",
                "target_handle_id": "8a4a7efd-0e18-43ed-ba32-803e22e3ba0a",
                "output_id": "6e7eeaa5-9559-4ae3-8606-e52ead5805a5",
                "output_type": "STRING",
                "node_input_id": "796b4a0b-da10-403a-acc3-8ebd3ebd3667",
            },
            "inputs": [
                {
                    "id": "796b4a0b-da10-403a-acc3-8ebd3ebd3667",
                    "key": "node_input",
                    "value": {
                        "rules": [
                            {
                                "type": "NODE_OUTPUT",
                                "data": {
                                    "node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                    "output_id": "f305c61e-6e8f-4cea-a53e-92656136b545",
                                },
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "display_data": {"position": {"x": 400.0, "y": 75.0}},
            "base": {
                "name": "FinalOutputNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
            },
            "definition": None,
        },
        final_output_node_input,
    )

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "26003bf1-b5a3-41f4-a419-c7ef5a7595df",
                "source_node_id": "32684932-7c7c-4b1c-aed2-553de29bf3f7",
                "source_handle_id": "e4136ee4-a51a-4ca3-9a3a-aa96f5de2347",
                "target_node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                "target_handle_id": "6492efcf-4437-4af1-9ad7-269795ccb27a",
                "type": "DEFAULT",
            },
            {
                "id": "b0a57a5f-a1e4-4dc9-85dd-946f08304738",
                "source_node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                "source_handle_id": "1e739e86-a285-4438-9725-a152c15a63e3",
                "target_node_id": "ca8bb585-c9a8-4bf7-bf9d-534b600fe23b",
                "target_handle_id": "8a4a7efd-0e18-43ed-ba32-803e22e3ba0a",
                "type": "DEFAULT",
            },
            {
                "id": "e4366583-94a5-40b0-9b6f-1e965695b1fe",
                "source_node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                "source_handle_id": "1e739e86-a285-4438-9725-a152c15a63e3",
                "target_node_id": "27fdaa45-b8ce-464d-be50-cf71cc56bc10",
                "target_handle_id": "e7a09eb2-c9fb-4d57-b436-9cd9384c8960",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicDefaultStateWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_default_state",
            "workflow",
        ],
    }



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value_request import ArrayVellumValueRequest
import typing
import pydantic
from .named_test_case_variable_value_request import NamedTestCaseVariableValueRequest
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class ReplaceTestSuiteTestCaseRequest(UniversalBaseModel):
    """
    Information about the Test Case to replace
    """

    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The Vellum-generated ID of the Test Case whose data you'd like to replace. Must specify either this or external_id.
    """

    external_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The ID that was originally provided upon Test Case creation that uniquely identifies the Test Case whose data you'd like to replace. Must specify either this of id.
    """

    label: typing.Optional[str] = pydantic.Field(default=None)
    """
    A human-readable label used to convey the intention of this Test Case
    """

    input_values: typing.List[NamedTestCaseVariableValueRequest] = pydantic.Field()
    """
    Values for each of the Test Case's input variables
    """

    evaluation_values: typing.List[NamedTestCaseVariableValueRequest] = pydantic.Field()
    """
    Values for each of the Test Case's evaluation variables
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .initiated_execute_prompt_event import InitiatedExecutePromptEvent
from .streaming_execute_prompt_event import StreamingExecutePromptEvent
from .fulfilled_execute_prompt_event import FulfilledExecutePromptEvent
from .rejected_execute_prompt_event import RejectedExecutePromptEvent

ExecutePromptEvent = typing.Union[
    InitiatedExecutePromptEvent, StreamingExecutePromptEvent, FulfilledExecutePromptEvent, RejectedExecutePromptEvent
]



from uuid import UUID
from typing import Generic, Optional, TypeVar

from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes import SubworkflowDeploymentNode
from vellum.workflows.types.core import JsonObject
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext

_SubworkflowDeploymentNodeType = TypeVar("_SubworkflowDeploymentNodeType", bound=SubworkflowDeploymentNode)


class BaseSubworkflowDeploymentNodeDisplay(
    BaseNodeDisplay[_SubworkflowDeploymentNodeType], Generic[_SubworkflowDeploymentNodeType]
):
    __serializable_inputs__ = {SubworkflowDeploymentNode.subworkflow_inputs}

    def serialize(
        self, display_context: WorkflowDisplayContext, error_output_id: Optional[UUID] = None, **_kwargs
    ) -> JsonObject:
        node = self._node
        node_id = self.node_id

        subworkflow_inputs = raise_if_descriptor(node.subworkflow_inputs)

        if isinstance(subworkflow_inputs, BaseInputs):
            input_items = [(input_descriptor.name, input_value) for input_descriptor, input_value in subworkflow_inputs]
        else:
            input_items = list(subworkflow_inputs.items())

        node_inputs = [
            create_node_input(
                node_id=node_id,
                input_name=variable_name,
                value=variable_value,
                display_context=display_context,
                input_id=self.node_input_ids_by_name.get(
                    f"{SubworkflowDeploymentNode.subworkflow_inputs.name}.{variable_name}"
                )
                or self.node_input_ids_by_name.get(variable_name),
            )
            for variable_name, variable_value in input_items
        ]

        deployment = display_context.client.workflow_deployments.retrieve(
            id=str(raise_if_descriptor(node.deployment)),
        )

        return {
            "id": str(node_id),
            "type": "SUBWORKFLOW",
            "inputs": [node_input.dict() for node_input in node_inputs],
            "data": {
                "label": self.label,
                "error_output_id": str(error_output_id) if error_output_id else None,
                "source_handle_id": str(self.get_source_handle_id(display_context.port_displays)),
                "target_handle_id": str(self.get_target_handle_id()),
                "variant": "DEPLOYMENT",
                "workflow_deployment_id": str(deployment.id),
                "release_tag": raise_if_descriptor(node.release_tag),
            },
            "display_data": self.get_display_data().dict(),
            "base": self.get_base().dict(),
            "definition": self.get_definition().dict(),
            "ports": self.serialize_ports(display_context),
        }



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .generate_result_data import GenerateResultData
import pydantic
from .generate_result_error import GenerateResultError
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class GenerateResult(UniversalBaseModel):
    data: typing.Optional[GenerateResultData] = pydantic.Field(default=None)
    """
    An object containing the resulting generation. This key will be absent if the LLM provider experienced an error.
    """

    error: typing.Optional[GenerateResultError] = pydantic.Field(default=None)
    """
    An object containing details about the error that occurred. This key will be absent if the LLM provider did not experience an error.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .test_suite_run_execution_string_output import TestSuiteRunExecutionStringOutput
from .test_suite_run_execution_number_output import TestSuiteRunExecutionNumberOutput
from .test_suite_run_execution_json_output import TestSuiteRunExecutionJsonOutput
from .test_suite_run_execution_chat_history_output import TestSuiteRunExecutionChatHistoryOutput
from .test_suite_run_execution_search_results_output import TestSuiteRunExecutionSearchResultsOutput
from .test_suite_run_execution_error_output import TestSuiteRunExecutionErrorOutput
from .test_suite_run_execution_function_call_output import TestSuiteRunExecutionFunctionCallOutput
from .test_suite_run_execution_array_output import TestSuiteRunExecutionArrayOutput

TestSuiteRunExecutionOutput = typing.Union[
    TestSuiteRunExecutionStringOutput,
    TestSuiteRunExecutionNumberOutput,
    TestSuiteRunExecutionJsonOutput,
    TestSuiteRunExecutionChatHistoryOutput,
    TestSuiteRunExecutionSearchResultsOutput,
    TestSuiteRunExecutionErrorOutput,
    TestSuiteRunExecutionFunctionCallOutput,
    TestSuiteRunExecutionArrayOutput,
]



from dataclasses import dataclass
from uuid import UUID
from typing import Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union, cast

from vellum import (
    MetadataFilterConfigRequest,
    VellumValueLogicalConditionGroupRequest,
    VellumValueLogicalConditionRequest,
)
from vellum.workflows.nodes.displayable.bases.types import MetadataLogicalCondition, MetadataLogicalConditionGroup
from vellum.workflows.nodes.displayable.search_node import SearchNode
from vellum.workflows.references import OutputReference
from vellum.workflows.types.core import JsonArray, JsonObject
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.vellum import InputVariablePointer
from vellum_ee.workflows.display.vellum import NodeInput

_SearchNodeType = TypeVar("_SearchNodeType", bound=SearchNode)


@dataclass
class VariableIdMap:
    id: Optional[UUID]
    lhs: Optional["VariableIdMap"]
    rhs: Optional["VariableIdMap"]


class BaseSearchNodeDisplay(BaseNodeDisplay[_SearchNodeType], Generic[_SearchNodeType]):
    # A mapping between the id of the operand (e.g. "lhs_variable_id" or "rhs_variable_id") and the id of the node input
    # that the operand is pointing to.
    metadata_filter_input_id_by_operand_id: Dict[UUID, UUID] = {}
    __serializable_inputs__ = {
        SearchNode.query,
        SearchNode.document_index,
        SearchNode.weights,
        SearchNode.chunk_separator,
        SearchNode.limit,
        SearchNode.result_merging,
    }

    def serialize(
        self, display_context: WorkflowDisplayContext, error_output_id: Optional[UUID] = None, **_kwargs
    ) -> JsonObject:
        node = self._node
        node_id = self.node_id
        node_inputs = self._generate_search_node_inputs(node_id, node, display_context)

        _, results_output_display = display_context.global_node_output_displays[
            cast(OutputReference, node.Outputs.results)
        ]
        _, text_output_display = display_context.global_node_output_displays[cast(OutputReference, node.Outputs.text)]

        return {
            "id": str(node_id),
            "type": "SEARCH",
            "inputs": [node_input.dict() for node_input in node_inputs.values()],
            "data": {
                "label": self.label,
                "results_output_id": str(results_output_display.id),
                "text_output_id": str(text_output_display.id),
                "error_output_id": str(error_output_id) if error_output_id else None,
                "source_handle_id": str(self.get_source_handle_id(display_context.port_displays)),
                "target_handle_id": str(self.get_target_handle_id()),
                "query_node_input_id": str(node_inputs["query"].id),
                "document_index_node_input_id": str(node_inputs["document_index_id"].id),
                "weights_node_input_id": str(node_inputs["weights"].id),
                "limit_node_input_id": str(node_inputs["limit"].id),
                "separator_node_input_id": str(node_inputs["separator"].id),
                "result_merging_enabled_node_input_id": str(node_inputs["result_merging_enabled"].id),
                "external_id_filters_node_input_id": str(node_inputs["external_id_filters"].id),
                "metadata_filters_node_input_id": str(node_inputs["metadata_filters"].id),
            },
            "display_data": self.get_display_data().dict(),
            "base": self.get_base().dict(),
            "definition": self.get_definition().dict(),
            "ports": self.serialize_ports(display_context),
        }

    def _generate_search_node_inputs(
        self,
        node_id: UUID,
        node: Type[SearchNode],
        display_context: WorkflowDisplayContext,
    ) -> Dict[str, NodeInput]:
        node_inputs: Dict[str, NodeInput] = {}

        options = raise_if_descriptor(node.options)
        raw_filters = raise_if_descriptor(node.filters)
        filters = raw_filters if raw_filters else options.filters if options else None

        external_id_filters = filters.external_ids if filters else None

        raw_metadata_filters = filters.metadata if filters else None
        metadata_filters = None
        metadata_filters_node_inputs: list[NodeInput] = []
        if raw_metadata_filters:
            if isinstance(raw_metadata_filters, MetadataFilterConfigRequest):
                raise ValueError(
                    "MetadataFilterConfigRequest is deprecated. Please use VellumValueLogicalExpressionRequest instead."
                )
            metadata_filters, metadata_filters_node_inputs = self._serialize_logical_expression(
                raw_metadata_filters, display_context=display_context
            )

        raw_result_merging = raise_if_descriptor(node.result_merging)
        result_merging = raw_result_merging if raw_result_merging else options.result_merging if options else None
        result_merging_enabled = True if result_merging and result_merging.enabled else False

        raw_weights = raise_if_descriptor(node.weights)
        weights = raw_weights if raw_weights is not None else options.weights if options is not None else None

        raw_limit = raise_if_descriptor(node.limit)
        limit = raw_limit if raw_limit is not None else options.limit if options is not None else None

        node_input_names_and_values = [
            ("query", node.query),
            ("document_index_id", node.document_index),
            ("weights", weights.dict() if weights else None),
            ("limit", limit),
            ("separator", raise_if_descriptor(node.chunk_separator)),
            (
                "result_merging_enabled",
                ("True" if result_merging_enabled else "False"),
            ),
            ("external_id_filters", external_id_filters),
            ("metadata_filters", metadata_filters),
        ]

        for node_input_name, node_input_value in node_input_names_and_values:
            node_input = create_node_input(
                node_id,
                node_input_name,
                node_input_value,
                display_context,
                input_id=self.node_input_ids_by_name.get(node_input_name),
            )
            node_inputs[node_input_name] = node_input

        for node_input in metadata_filters_node_inputs:
            node_inputs[node_input.key] = node_input

        return node_inputs

    def _serialize_logical_expression(
        self,
        logical_expression: Union[
            VellumValueLogicalConditionGroupRequest,
            VellumValueLogicalConditionRequest,
            MetadataLogicalConditionGroup,
            MetadataLogicalCondition,
        ],
        display_context: WorkflowDisplayContext,
        path: List[int] = [],
    ) -> Tuple[JsonObject, List[NodeInput]]:
        if isinstance(logical_expression, VellumValueLogicalConditionGroupRequest):
            conditions: JsonArray = []
            variables = []
            for idx, condition in enumerate(logical_expression.conditions):
                serialized_condition, serialized_variables = self._serialize_logical_expression(
                    condition, display_context=display_context, path=path + [idx]
                )
                conditions.append(serialized_condition)
                variables.extend(serialized_variables)

            return (
                {
                    "type": "LOGICAL_CONDITION_GROUP",
                    "combinator": logical_expression.combinator,
                    "conditions": conditions,
                    "negated": logical_expression.negated,
                },
                variables,
            )
        elif isinstance(logical_expression, VellumValueLogicalConditionRequest):
            lhs_variable_id = logical_expression.lhs_variable.value
            if not isinstance(lhs_variable_id, str):
                raise TypeError(f"Expected lhs_variable_id to be a string, got {type(lhs_variable_id)}")

            rhs_variable_id = logical_expression.rhs_variable.value
            if not isinstance(rhs_variable_id, str):
                raise TypeError(f"Expected rhs_variable_id to be a string, got {type(rhs_variable_id)}")

            lhs_query_input_id: UUID = self.metadata_filter_input_id_by_operand_id.get(
                UUID(lhs_variable_id)
            ) or uuid4_from_hash(f"{self.node_id}|lhs|{hash(tuple(path))}")
            rhs_query_input_id: UUID = self.metadata_filter_input_id_by_operand_id.get(
                UUID(rhs_variable_id)
            ) or uuid4_from_hash(f"{self.node_id}|rhs|{hash(tuple(path))}")

            return (
                {
                    "type": "LOGICAL_CONDITION",
                    "lhs_variable_id": str(lhs_variable_id),
                    "operator": logical_expression.operator,
                    "rhs_variable_id": str(rhs_variable_id),
                },
                [
                    create_node_input(
                        self.node_id,
                        f"vellum-query-builder-variable-{lhs_variable_id}",
                        str(lhs_query_input_id),
                        display_context,
                        input_id=UUID(lhs_variable_id),
                        pointer_type=InputVariablePointer,
                    ),
                    create_node_input(
                        self.node_id,
                        f"vellum-query-builder-variable-{rhs_variable_id}",
                        str(rhs_query_input_id),
                        display_context,
                        input_id=UUID(rhs_variable_id),
                        pointer_type=InputVariablePointer,
                    ),
                ],
            )

        elif isinstance(logical_expression, MetadataLogicalConditionGroup):
            conditions = []
            variables = []
            for idx, metadata_condition in enumerate(logical_expression.conditions):
                serialized_condition, serialized_variables = self._serialize_logical_expression(
                    metadata_condition, display_context=display_context, path=path + [idx]
                )
                conditions.append(serialized_condition)
                variables.extend(serialized_variables)

            return (
                {
                    "type": "LOGICAL_CONDITION_GROUP",
                    "combinator": logical_expression.combinator,
                    "conditions": conditions,
                    "negated": logical_expression.negated,
                },
                variables,
            )

        elif isinstance(logical_expression, MetadataLogicalCondition):
            lhs_variable = logical_expression.lhs_variable
            rhs_variable = logical_expression.rhs_variable

            lhs_query_input_id = uuid4_from_hash(f"{self.node_id}|lhs|{hash(tuple(path))}")
            rhs_query_input_id = uuid4_from_hash(f"{self.node_id}|rhs|{hash(tuple(path))}")

            return (
                {
                    "type": "LOGICAL_CONDITION",
                    "lhs_variable_id": str(lhs_query_input_id),
                    "operator": logical_expression.operator,
                    "rhs_variable_id": str(rhs_query_input_id),
                },
                [
                    create_node_input(
                        self.node_id,
                        f"vellum-query-builder-variable-{lhs_query_input_id}",
                        lhs_variable,
                        display_context,
                        input_id=lhs_query_input_id,
                    ),
                    create_node_input(
                        self.node_id,
                        f"vellum-query-builder-variable-{rhs_query_input_id}",
                        rhs_variable,
                        display_context,
                        input_id=rhs_query_input_id,
                    ),
                ],
            )
        else:
            raise ValueError(f"Unsupported logical expression type: {type(logical_expression)}")



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from .raw_client import RawWorkflowsClient
from ...core.request_options import RequestOptions
from ...types.workflow_push_exec_config import WorkflowPushExecConfig
from ...types.workflow_push_deployment_config_request import WorkflowPushDeploymentConfigRequest
from ... import core
from ...types.workflow_push_response import WorkflowPushResponse
from ...core.client_wrapper import AsyncClientWrapper
from .raw_client import AsyncRawWorkflowsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class WorkflowsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawWorkflowsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawWorkflowsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawWorkflowsClient
        """
        return self._raw_client

    def pull(
        self,
        id: str,
        *,
        exclude_code: typing.Optional[bool] = None,
        exclude_display: typing.Optional[bool] = None,
        include_json: typing.Optional[bool] = None,
        include_sandbox: typing.Optional[bool] = None,
        strict: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[bytes]:
        """
        Parameters
        ----------
        id : str
            The ID of the Workflow to pull from

        exclude_code : typing.Optional[bool]

        exclude_display : typing.Optional[bool]

        include_json : typing.Optional[bool]

        include_sandbox : typing.Optional[bool]

        strict : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[bytes]

        """
        with self._raw_client.pull(
            id,
            exclude_code=exclude_code,
            exclude_display=exclude_display,
            include_json=include_json,
            include_sandbox=include_sandbox,
            strict=strict,
            request_options=request_options,
        ) as r:
            yield from r.data

    def push(
        self,
        *,
        exec_config: WorkflowPushExecConfig,
        workflow_sandbox_id: typing.Optional[str] = OMIT,
        deployment_config: typing.Optional[WorkflowPushDeploymentConfigRequest] = OMIT,
        artifact: typing.Optional[core.File] = OMIT,
        dry_run: typing.Optional[bool] = OMIT,
        strict: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowPushResponse:
        """
        Parameters
        ----------
        exec_config : WorkflowPushExecConfig
            The execution configuration of the workflow.

        workflow_sandbox_id : typing.Optional[str]

        deployment_config : typing.Optional[WorkflowPushDeploymentConfigRequest]

        artifact : typing.Optional[core.File]
            See core.File for more documentation

        dry_run : typing.Optional[bool]

        strict : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowPushResponse

        """
        response = self._raw_client.push(
            exec_config=exec_config,
            workflow_sandbox_id=workflow_sandbox_id,
            deployment_config=deployment_config,
            artifact=artifact,
            dry_run=dry_run,
            strict=strict,
            request_options=request_options,
        )
        return response.data


class AsyncWorkflowsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawWorkflowsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawWorkflowsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawWorkflowsClient
        """
        return self._raw_client

    async def pull(
        self,
        id: str,
        *,
        exclude_code: typing.Optional[bool] = None,
        exclude_display: typing.Optional[bool] = None,
        include_json: typing.Optional[bool] = None,
        include_sandbox: typing.Optional[bool] = None,
        strict: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[bytes]:
        """
        Parameters
        ----------
        id : str
            The ID of the Workflow to pull from

        exclude_code : typing.Optional[bool]

        exclude_display : typing.Optional[bool]

        include_json : typing.Optional[bool]

        include_sandbox : typing.Optional[bool]

        strict : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[bytes]

        """
        async with self._raw_client.pull(
            id,
            exclude_code=exclude_code,
            exclude_display=exclude_display,
            include_json=include_json,
            include_sandbox=include_sandbox,
            strict=strict,
            request_options=request_options,
        ) as r:
            async for data in r.data:
                yield data

    async def push(
        self,
        *,
        exec_config: WorkflowPushExecConfig,
        workflow_sandbox_id: typing.Optional[str] = OMIT,
        deployment_config: typing.Optional[WorkflowPushDeploymentConfigRequest] = OMIT,
        artifact: typing.Optional[core.File] = OMIT,
        dry_run: typing.Optional[bool] = OMIT,
        strict: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowPushResponse:
        """
        Parameters
        ----------
        exec_config : WorkflowPushExecConfig
            The execution configuration of the workflow.

        workflow_sandbox_id : typing.Optional[str]

        deployment_config : typing.Optional[WorkflowPushDeploymentConfigRequest]

        artifact : typing.Optional[core.File]
            See core.File for more documentation

        dry_run : typing.Optional[bool]

        strict : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowPushResponse

        """
        response = await self._raw_client.push(
            exec_config=exec_config,
            workflow_sandbox_id=workflow_sandbox_id,
            deployment_config=deployment_config,
            artifact=artifact,
            dry_run=dry_run,
            strict=strict,
            request_options=request_options,
        )
        return response.data



from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_tool_calling_node_with_composio_tool.workflow import (
    BasicToolCallingNodeWithComposioToolWorkflow,
)


def test_serialize_workflow():
    # GIVEN a Workflow that uses a tool calling node with a composio tool
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicToolCallingNodeWithComposioToolWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2

    # Find the text and chat_history outputs
    text_output = next(var for var in output_variables if var["key"] == "text")
    chat_history_output = next(var for var in output_variables if var["key"] == "chat_history")

    assert text_output["type"] == "STRING"
    assert chat_history_output["type"] == "CHAT_HISTORY"

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    tool_calling_node = workflow_raw_data["nodes"][1]

    # AND the tool calling node should have the composio tool properly serialized
    functions_attribute = next(attr for attr in tool_calling_node["attributes"] if attr["name"] == "functions")
    assert functions_attribute["value"]["type"] == "CONSTANT_VALUE"
    assert functions_attribute["value"]["value"]["type"] == "JSON"

    functions_list = functions_attribute["value"]["value"]["value"]
    assert len(functions_list) == 1

    composio_function = functions_list[0]
    assert composio_function == {
        "type": "COMPOSIO",
        "toolkit": "GITHUB",
        "action": "GITHUB_CREATE_AN_ISSUE",
        "description": "Create a new issue in a GitHub repository",
        "user_id": None,
    }

    # AND the rest of the node structure should be correct
    assert tool_calling_node["type"] == "GENERIC"
    assert tool_calling_node["base"]["name"] == "ToolCallingNode"
    assert tool_calling_node["definition"]["name"] == "ComposioToolCallingNode"

    # AND the blocks should be properly serialized
    blocks_attribute = next(attr for attr in tool_calling_node["attributes"] if attr["name"] == "blocks")
    assert blocks_attribute["value"]["type"] == "CONSTANT_VALUE"
    blocks_list = blocks_attribute["value"]["value"]["value"]
    assert len(blocks_list) == 2
    assert blocks_list[0]["chat_role"] == "SYSTEM"
    assert blocks_list[1]["chat_role"] == "USER"

    # AND the prompt inputs should be properly serialized
    prompt_inputs_attribute = next(attr for attr in tool_calling_node["attributes"] if attr["name"] == "prompt_inputs")
    assert prompt_inputs_attribute["value"]["type"] == "DICTIONARY_REFERENCE"
    entries = prompt_inputs_attribute["value"]["entries"]
    assert len(entries) == 1
    assert entries[0]["key"] == "question"
    assert entries[0]["value"]["type"] == "WORKFLOW_INPUT"

    # AND the outputs should be correct
    outputs = tool_calling_node["outputs"]
    assert len(outputs) == 2
    assert outputs[0]["name"] == "text"
    assert outputs[0]["type"] == "STRING"
    assert outputs[1]["name"] == "chat_history"
    assert outputs[1]["type"] == "CHAT_HISTORY"



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value import ArrayVellumValue
import datetime as dt
import pydantic
import typing
from .vellum_variable import VellumVariable
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class DeploymentHistoryItem(UniversalBaseModel):
    id: str
    deployment_id: str
    timestamp: dt.datetime
    label: str = pydantic.Field()
    """
    A human-readable label for the deployment
    """

    name: str = pydantic.Field()
    """
    A name that uniquely identifies this deployment within its workspace
    """

    input_variables: typing.List[VellumVariable]
    description: typing.Optional[str] = pydantic.Field(default=None)
    """
    A human-readable description of the deployment
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from .raw_client import RawPromptsClient
from ...core.request_options import RequestOptions
from ...types.prompt_exec_config import PromptExecConfig
from ...types.prompt_push_response import PromptPushResponse
from ...core.client_wrapper import AsyncClientWrapper
from .raw_client import AsyncRawPromptsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class PromptsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawPromptsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawPromptsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawPromptsClient
        """
        return self._raw_client

    def pull(
        self,
        id: str,
        *,
        prompt_variant_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptExecConfig:
        """
        Used to pull the definition of a Prompt from Vellum.

        Parameters
        ----------
        id : str
            The ID of the Prompt to pull from. Prompt Sandbox IDs are currently supported.

        prompt_variant_id : typing.Optional[str]
            The ID of the Prompt Variant within a Prompt Sandbox to pull. Must be included if providing the ID of a Prompt Sandbox.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptExecConfig


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.prompts.pull(
            id="id",
        )
        """
        response = self._raw_client.pull(id, prompt_variant_id=prompt_variant_id, request_options=request_options)
        return response.data

    def push(
        self,
        *,
        exec_config: PromptExecConfig,
        prompt_variant_id: typing.Optional[str] = OMIT,
        prompt_variant_label: typing.Optional[str] = OMIT,
        prompt_sandbox_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptPushResponse:
        """
        Used to push updates to a Prompt in Vellum.

        Parameters
        ----------
        exec_config : PromptExecConfig

        prompt_variant_id : typing.Optional[str]
            If specified, an existing Prompt Variant by the provided ID will be updated. Otherwise, a new Prompt Variant will be created and an ID generated.

        prompt_variant_label : typing.Optional[str]
            If provided, then the created/updated Prompt Variant will have this label.

        prompt_sandbox_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptPushResponse


        Examples
        --------
        from vellum import (
            JinjaPromptBlock,
            PromptExecConfig,
            PromptParameters,
            Vellum,
            VellumVariable,
        )

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.prompts.push(
            exec_config=PromptExecConfig(
                ml_model="ml_model",
                input_variables=[
                    VellumVariable(
                        id="x",
                        key="key",
                        type="STRING",
                    ),
                    VellumVariable(
                        id="x",
                        key="key",
                        type="STRING",
                    ),
                ],
                parameters=PromptParameters(),
                blocks=[
                    JinjaPromptBlock(
                        template="template",
                    ),
                    JinjaPromptBlock(
                        template="template",
                    ),
                ],
            ),
        )
        """
        response = self._raw_client.push(
            exec_config=exec_config,
            prompt_variant_id=prompt_variant_id,
            prompt_variant_label=prompt_variant_label,
            prompt_sandbox_id=prompt_sandbox_id,
            request_options=request_options,
        )
        return response.data


class AsyncPromptsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawPromptsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawPromptsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawPromptsClient
        """
        return self._raw_client

    async def pull(
        self,
        id: str,
        *,
        prompt_variant_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptExecConfig:
        """
        Used to pull the definition of a Prompt from Vellum.

        Parameters
        ----------
        id : str
            The ID of the Prompt to pull from. Prompt Sandbox IDs are currently supported.

        prompt_variant_id : typing.Optional[str]
            The ID of the Prompt Variant within a Prompt Sandbox to pull. Must be included if providing the ID of a Prompt Sandbox.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptExecConfig


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.prompts.pull(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.pull(id, prompt_variant_id=prompt_variant_id, request_options=request_options)
        return response.data

    async def push(
        self,
        *,
        exec_config: PromptExecConfig,
        prompt_variant_id: typing.Optional[str] = OMIT,
        prompt_variant_label: typing.Optional[str] = OMIT,
        prompt_sandbox_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptPushResponse:
        """
        Used to push updates to a Prompt in Vellum.

        Parameters
        ----------
        exec_config : PromptExecConfig

        prompt_variant_id : typing.Optional[str]
            If specified, an existing Prompt Variant by the provided ID will be updated. Otherwise, a new Prompt Variant will be created and an ID generated.

        prompt_variant_label : typing.Optional[str]
            If provided, then the created/updated Prompt Variant will have this label.

        prompt_sandbox_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptPushResponse


        Examples
        --------
        import asyncio

        from vellum import (
            AsyncVellum,
            JinjaPromptBlock,
            PromptExecConfig,
            PromptParameters,
            VellumVariable,
        )

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.prompts.push(
                exec_config=PromptExecConfig(
                    ml_model="ml_model",
                    input_variables=[
                        VellumVariable(
                            id="x",
                            key="key",
                            type="STRING",
                        ),
                        VellumVariable(
                            id="x",
                            key="key",
                            type="STRING",
                        ),
                    ],
                    parameters=PromptParameters(),
                    blocks=[
                        JinjaPromptBlock(
                            template="template",
                        ),
                        JinjaPromptBlock(
                            template="template",
                        ),
                    ],
                ),
            )


        asyncio.run(main())
        """
        response = await self._raw_client.push(
            exec_config=exec_config,
            prompt_variant_id=prompt_variant_id,
            prompt_variant_label=prompt_variant_label,
            prompt_sandbox_id=prompt_sandbox_id,
            request_options=request_options,
        )
        return response.data



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations
from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .metadata_filter_rule_combinator import MetadataFilterRuleCombinator
from .logical_operator import LogicalOperator
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic
from ..core.pydantic_utilities import update_forward_refs


class MetadataFilterRuleRequest(UniversalBaseModel):
    combinator: typing.Optional[MetadataFilterRuleCombinator] = None
    negated: typing.Optional[bool] = None
    rules: typing.Optional[typing.List["MetadataFilterRuleRequest"]] = None
    field: typing.Optional[str] = None
    operator: typing.Optional[LogicalOperator] = None
    value: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


update_forward_refs(MetadataFilterRuleRequest)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value_request import ArrayVellumValueRequest
import typing
from .vellum_value_request import VellumValueRequest
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class NamedTestCaseArrayVariableValueRequest(UniversalBaseModel):
    """
    Named Test Case value that is of type ARRAY
    """

    type: typing.Literal["ARRAY"] = "ARRAY"
    value: typing.Optional[typing.List[VellumValueRequest]] = None
    name: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .code_execution_node_string_result import CodeExecutionNodeStringResult
from .code_execution_node_number_result import CodeExecutionNodeNumberResult
from .code_execution_node_json_result import CodeExecutionNodeJsonResult
from .code_execution_node_chat_history_result import CodeExecutionNodeChatHistoryResult
from .code_execution_node_search_results_result import CodeExecutionNodeSearchResultsResult
from .code_execution_node_error_result import CodeExecutionNodeErrorResult
from .code_execution_node_array_result import CodeExecutionNodeArrayResult
from .code_execution_node_function_call_result import CodeExecutionNodeFunctionCallResult

CodeExecutionNodeResultOutput = typing.Union[
    CodeExecutionNodeStringResult,
    CodeExecutionNodeNumberResult,
    CodeExecutionNodeJsonResult,
    CodeExecutionNodeChatHistoryResult,
    CodeExecutionNodeSearchResultsResult,
    CodeExecutionNodeErrorResult,
    CodeExecutionNodeArrayResult,
    CodeExecutionNodeFunctionCallResult,
]



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_error_node.workflow import BasicErrorNodeWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow with an error node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicErrorNodeWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "5d9edd44-b35b-4bad-ad51-ccdfe8185ff5",
                "key": "threshold",
                "type": "NUMBER",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            }
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "04c5c6be-f5e1-41b8-b668-39e179790d9e",
                "key": "final_value",
                "type": "NUMBER",
            }
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 4
    assert len(workflow_raw_data["nodes"]) == 5

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "10e90662-e998-421d-a5c9-ec16e37a8de1",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "7d86498b-84ed-4feb-8e62-2188058c2c4e",
        },
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    error_node, error_index = next(
        (
            (node, index)
            for index, node in enumerate(workflow_raw_data["nodes"])
            if node.get("data", {}).get("label") == "Fail Node"
        ),
        (None, None),
    )
    assert not DeepDiff(
        {
            "id": "5cf9c5e3-0eae-4daf-8d73-8b9536258eb9",
            "type": "ERROR",
            "inputs": [
                {
                    "id": "8e4c8d76-2e02-4d7e-a7bf-d71af392dc49",
                    "key": "error_source_input_id",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "Input threshold was too low"},
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "data": {
                "label": "Fail Node",
                "target_handle_id": "70c19f1c-309c-4a5d-ba65-664c0bb2fedf",
                "error_source_input_id": "8e4c8d76-2e02-4d7e-a7bf-d71af392dc49",
            },
            "display_data": {"position": {"x": 400.0, "y": 75.0}},
            "base": {
                "name": "ErrorNode",
                "module": ["vellum", "workflows", "nodes", "core", "error_node", "node"],
            },
            "definition": {
                "name": "FailNode",
                "module": ["tests", "workflows", "basic_error_node", "workflow"],
            },
        },
        error_node,
        ignore_order=True,
    )

    passthrough_nodes = [node for node in workflow_raw_data["nodes"] if node["type"] == "GENERIC"]
    assert len(passthrough_nodes) == 2

    terminal_node = workflow_raw_data["nodes"][-1]
    assert not DeepDiff(
        {
            "id": "e5fff999-80c7-4cbc-9d99-06c653f3ec77",
            "type": "TERMINAL",
            "data": {
                "label": "Final Output",
                "name": "final_value",
                "target_handle_id": "b070e9bc-e9b7-46d3-8f5b-0b646bd25cf0",
                "output_id": "04c5c6be-f5e1-41b8-b668-39e179790d9e",
                "output_type": "NUMBER",
                "node_input_id": "c191a5c1-8912-49ad-bf4b-c6d4d58ed482",
            },
            "inputs": [
                {
                    "id": "c191a5c1-8912-49ad-bf4b-c6d4d58ed482",
                    "key": "node_input",
                    "value": {
                        "rules": [
                            {
                                "type": "NODE_OUTPUT",
                                "data": {
                                    "node_id": "1eee9b4e-531f-45f2-a4b9-42207fac2c33",
                                    "output_id": "c6b017a4-25e9-4296-8d81-6aa4b3dad171",
                                },
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "display_data": {"position": {"x": 600.0, "y": -50.0}},
            "base": {
                "name": "FinalOutputNode",
                "module": [
                    "vellum",
                    "workflows",
                    "nodes",
                    "displayable",
                    "final_output_node",
                    "node",
                ],
            },
            "definition": None,
        },
        terminal_node,
    )



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import pydantic
import typing
from .indexing_state_enum import IndexingStateEnum
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class DocumentDocumentToDocumentIndex(UniversalBaseModel):
    """
    A detailed representation of the link between a Document and a Document Index it's a member of.
    """

    id: str = pydantic.Field()
    """
    Vellum-generated ID that uniquely identifies this link.
    """

    environment_document_index_id: str = pydantic.Field()
    """
    Vellum-generated ID that uniquely identifies the environment index this document is included in.
    """

    document_index_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    Vellum-generated ID that uniquely identifies the index this document is included in.
    """

    indexing_state: typing.Optional[IndexingStateEnum] = pydantic.Field(default=None)
    """
    An enum value representing where this document is along its indexing lifecycle for this index.
    
    * `AWAITING_PROCESSING` - Awaiting Processing
    * `QUEUED` - Queued
    * `INDEXING` - Indexing
    * `INDEXED` - Indexed
    * `FAILED` - Failed
    """

    extracted_text_file_url: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from abc import ABC, ABCMeta
from dataclasses import field
from functools import cached_property, reduce
import inspect
from types import MappingProxyType
from uuid import UUID, uuid4
from typing import Any, Dict, Generic, Iterator, Optional, Set, Tuple, Type, TypeVar, Union, cast, get_args

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.descriptors.utils import is_unresolved, resolve_value
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.events.node import NodeExecutionStreamingEvent
from vellum.workflows.exceptions import NodeException
from vellum.workflows.graph import Graph
from vellum.workflows.graph.graph import GraphTarget
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.outputs import BaseOutput, BaseOutputs
from vellum.workflows.ports.node_ports import NodePorts
from vellum.workflows.ports.port import Port
from vellum.workflows.references import ExternalInputReference
from vellum.workflows.references.execution_count import ExecutionCountReference
from vellum.workflows.references.node import NodeReference
from vellum.workflows.references.output import OutputReference
from vellum.workflows.state.base import BaseState
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.types.core import MergeBehavior
from vellum.workflows.types.generics import StateType
from vellum.workflows.types.utils import get_class_attr_names, get_original_base, infer_types
from vellum.workflows.utils.uuids import uuid4_from_hash


def _is_nested_class(nested: Any, parent: Type) -> bool:
    return (
        inspect.isclass(nested)
        # If a class is defined within a function, we don't consider it nested in the class defining that function
        # The example of this is a Subworkflow defined within TryNode.wrap()
        and (len(nested.__qualname__.split(".")) < 2 or nested.__qualname__.split(".")[-2] != "<locals>")
        and nested.__module__ == parent.__module__
        and (nested.__qualname__.startswith(parent.__name__) or nested.__qualname__.startswith(parent.__qualname__))
    ) or any(_is_nested_class(nested, base) for base in parent.__bases__)


def _is_annotated(cls: Type, name: str) -> bool:
    if name in cls.__annotations__:
        return True

    for base in cls.__bases__:
        if _is_annotated(base, name):
            return True

    return False


class BaseNodeMeta(ABCMeta):
    def __new__(mcs, name: str, bases: Tuple[Type, ...], dct: Dict[str, Any]) -> Any:
        if "Outputs" in dct:
            outputs_class = dct["Outputs"]
            if not any(issubclass(base, BaseOutputs) for base in outputs_class.__bases__):
                parent_outputs_class = next(
                    (base.Outputs for base in bases if hasattr(base, "Outputs")),
                    BaseOutputs,  # Default to BaseOutputs only if no parent has Outputs
                )

                # Filter out object from bases while preserving other inheritance
                filtered_bases = tuple(base for base in outputs_class.__bases__ if base is not object)

                dct["Outputs"] = type(
                    f"{name}.Outputs",
                    (parent_outputs_class,) + filtered_bases,
                    {**outputs_class.__dict__, "__module__": dct["__module__"]},
                )
        else:
            for base in reversed(bases):
                if hasattr(base, "Outputs"):
                    dct["Outputs"] = type(
                        f"{name}.Outputs",
                        (base.Outputs,),
                        {"__module__": dct["__module__"]},
                    )
                    break
            else:
                raise ValueError("Outputs class not found in base classes")

        if "Ports" in dct:
            dct["Ports"] = type(
                f"{name}.Ports",
                (NodePorts,),
                {**dct["Ports"].__dict__, "__module__": dct["__module__"]},
            )
        else:
            for base in reversed(bases):
                if issubclass(base, BaseNode):
                    ports_dct = {p.name: Port(default=p.default) for p in base.Ports}
                    ports_dct["__module__"] = dct["__module__"]
                    dct["Ports"] = type(f"{name}.Ports", (base.Ports,), ports_dct)
                    break

        if "Execution" not in dct:
            for base in reversed(bases):
                if issubclass(base, BaseNode):
                    dct["Execution"] = type(
                        f"{name}.Execution",
                        (base.Execution,),
                        {"__module__": dct["__module__"]},
                    )
                    break

        if "Trigger" not in dct:
            for base in reversed(bases):
                if issubclass(base, BaseNode):
                    trigger_dct = {
                        **base.Trigger.__dict__,
                        "__module__": dct["__module__"],
                    }
                    dct["Trigger"] = type(f"{name}.Trigger", (base.Trigger,), trigger_dct)
                    break

        cls = super().__new__(mcs, name, bases, dct)
        node_class = cast(Type["BaseNode"], cls)

        node_class.Outputs._node_class = node_class

        # Add cls to relevant nested classes, since python should've been doing this by default
        for port in node_class.Ports:
            port.node_class = node_class
            port.validate()

        node_class.Execution.node_class = node_class
        node_class.Trigger.node_class = node_class
        node_class.ExternalInputs.__parent_class__ = node_class
        node_class.__id__ = uuid4_from_hash(node_class.__qualname__)
        node_class.__output_ids__ = {
            ref.name: uuid4_from_hash(f"{node_class.__id__}|{ref.name}")
            for ref in node_class.Outputs
            if isinstance(ref, OutputReference)
        }
        return node_class

    @property
    def _localns(cls) -> Dict[str, Any]:
        from vellum.workflows.workflows.base import BaseWorkflow

        return {"BaseWorkflow": BaseWorkflow}

    def __getattribute__(cls, name: str) -> Any:
        if name.startswith("_"):
            return super().__getattribute__(name)

        try:
            attribute = super().__getattribute__(name)
        except AttributeError as e:
            if _is_annotated(cls, name):
                attribute = None
            else:
                raise e

        if (
            inspect.isfunction(attribute)
            or inspect.ismethod(attribute)
            or _is_nested_class(attribute, cls)
            or isinstance(attribute, (property, cached_property))
            or not issubclass(cls, BaseNode)
        ):
            return attribute

        types = infer_types(cls, name, cls._localns)
        return NodeReference(
            name=name,
            types=types,
            instance=attribute,
            node_class=cls,
        )

    def __rshift__(cls, other_cls: GraphTarget) -> Graph:
        if not issubclass(cls, BaseNode):
            raise ValueError("BaseNodeMeta can only be extended from subclasses of BaseNode")

        if not cls.Ports._default_port:
            raise ValueError("No default port found on node")

        if isinstance(other_cls, Graph) or isinstance(other_cls, set):
            return Graph.from_node(cls) >> other_cls

        return cls.Ports._default_port >> other_cls

    def __rrshift__(cls, other_cls: GraphTarget) -> Graph:
        if not issubclass(cls, BaseNode):
            raise ValueError("BaseNodeMeta can only be extended from subclasses of BaseNode")

        if not isinstance(other_cls, set):
            other_cls = {other_cls}

        return Graph.from_set(other_cls) >> cls

    def __repr__(self) -> str:
        return f"{self.__module__}.{self.__qualname__}"

    def __iter__(cls) -> Iterator[NodeReference]:
        # We iterate through the inheritance hierarchy to find all the OutputDescriptors attached to this Outputs class.
        # __mro__ is the method resolution order, which is the order in which base classes are resolved.
        yielded_attr_names: Set[str] = {"state"}

        for resolved_cls in cls.__mro__:
            attr_names = get_class_attr_names(resolved_cls)
            for attr_name in attr_names:
                if attr_name in yielded_attr_names:
                    continue

                attr_value = getattr(resolved_cls, attr_name, undefined)
                if not isinstance(attr_value, NodeReference):
                    continue

                yield attr_value
                yielded_attr_names.add(attr_name)


class _BaseNodeTriggerMeta(type):
    def __eq__(self, other: Any) -> bool:
        """
        We need to include custom eq logic to prevent infinite loops during ipython reloading.
        """

        if not isinstance(other, _BaseNodeTriggerMeta):
            return False

        if not self.__name__.endswith(".Trigger") or not other.__name__.endswith(".Trigger"):
            return super().__eq__(other)

        self_trigger_class = cast(Type["BaseNode.Trigger"], self)
        other_trigger_class = cast(Type["BaseNode.Trigger"], other)

        return self_trigger_class.node_class.__name__ == other_trigger_class.node_class.__name__


class _BaseNodeExecutionMeta(type):
    def __getattribute__(cls, name: str) -> Any:
        if name == "count" and issubclass(cls, BaseNode.Execution):
            return ExecutionCountReference(cls.node_class)

        return super().__getattribute__(name)

    def __eq__(self, other: Any) -> bool:
        """
        We need to include custom eq logic to prevent infinite loops during ipython reloading.
        """

        if not isinstance(other, _BaseNodeExecutionMeta):
            return False

        if not self.__name__.endswith(".Execution") or not other.__name__.endswith(".Execution"):
            return super().__eq__(other)

        self_execution_class = cast(Type["BaseNode.Execution"], self)
        other_execution_class = cast(Type["BaseNode.Execution"], other)

        return self_execution_class.node_class.__name__ == other_execution_class.node_class.__name__


NodeRunResponse = Union[BaseOutputs, Iterator[BaseOutput]]


class BaseNode(Generic[StateType], ABC, metaclass=BaseNodeMeta):
    __id__: UUID = uuid4_from_hash(__qualname__)
    __output_ids__: Dict[str, UUID] = {}
    state: StateType
    _context: WorkflowContext
    _inputs: MappingProxyType[NodeReference, Any]

    class ExternalInputs(BaseInputs):
        __descriptor_class__ = ExternalInputReference

    # TODO: Consider using metaclasses to prevent the need for users to specify that the
    #   "Outputs" class inherits from "BaseOutputs" and do so automatically.
    #   https://app.shortcut.com/vellum/story/4008/auto-inherit-basenodeoutputs-in-outputs-classes
    class Outputs(BaseOutputs):
        _node_class: Type["BaseNode"] = field(init=False)

    class Ports(NodePorts):
        default = Port(default=True)

    class Trigger(metaclass=_BaseNodeTriggerMeta):
        node_class: Type["BaseNode"]
        merge_behavior = MergeBehavior.AWAIT_ATTRIBUTES

        @classmethod
        def should_initiate(
            cls,
            state: StateType,
            dependencies: Set["Type[BaseNode]"],
            node_span_id: UUID,
        ) -> bool:
            """
            Determines whether a Node's execution should be initiated. Override this method to define custom
            trigger criteria.
            """
            if state.meta.node_execution_cache.is_node_execution_initiated(cls.node_class, node_span_id):
                return False

            if cls.merge_behavior == MergeBehavior.AWAIT_ATTRIBUTES:
                for descriptor in cls.node_class:
                    if not descriptor.instance:
                        continue

                    try:
                        resolved_value = resolve_value(descriptor.instance, state, path=descriptor.name)
                    except InvalidExpressionException as e:
                        raise NodeException(
                            message=str(e),
                            code=WorkflowErrorCode.INVALID_INPUTS,
                        ) from e

                    if is_unresolved(resolved_value):
                        return False

                return True

            if cls.merge_behavior == MergeBehavior.AWAIT_ANY:
                return True

            if cls.merge_behavior == MergeBehavior.AWAIT_ALL:
                """
                A node utilizing an AWAIT_ALL merge strategy will only be considered ready
                when all of its dependencies have invoked this node.
                """
                # Check if all dependencies have invoked this node
                dependencies_invoked = state.meta.node_execution_cache._dependencies_invoked.get(node_span_id, set())
                node_classes_invoked = {
                    state.meta.node_execution_cache.__node_execution_lookup__[dep]
                    for dep in dependencies_invoked
                    if dep in state.meta.node_execution_cache.__node_execution_lookup__
                }
                if len(node_classes_invoked) != len(dependencies):
                    return False

                all_deps_invoked = all(dep in node_classes_invoked for dep in dependencies)
                return all_deps_invoked

            raise NodeException(
                message="Invalid Trigger Node Specification",
                code=WorkflowErrorCode.INVALID_INPUTS,
            )

        @classmethod
        def _queue_node_execution(
            cls, state: StateType, dependencies: Set["Type[BaseNode]"], invoked_by: Optional[UUID] = None
        ) -> UUID:
            """
            Queues a future execution of a node, returning the span id of the execution.

            We may combine this into the should_initiate method, but we'll keep it separate for now to avoid
            breaking changes until the 0.15.0 release.
            """

            execution_id = uuid4()
            if not invoked_by:
                return execution_id

            if invoked_by not in state.meta.node_execution_cache.__node_execution_lookup__:
                return execution_id

            if cls.merge_behavior not in {MergeBehavior.AWAIT_ANY, MergeBehavior.AWAIT_ALL}:
                # Keep track of the dependencies that have invoked this node
                # This would be needed while climbing the history in the loop
                state.meta.node_execution_cache._dependencies_invoked[execution_id].add(invoked_by)
                return execution_id

            # For AWAIT_ANY in workflows, we need to detect if the node is in a loop
            # If the node is in a loop, we can trigger the node again
            in_loop = False
            if cls.merge_behavior == MergeBehavior.AWAIT_ANY:
                # Get the latest fulfilled execution ID of current node
                fulfilled_stack = state.meta.node_execution_cache._node_executions_fulfilled[cls.node_class]
                current_latest_fulfilled_id = fulfilled_stack.peek() if not fulfilled_stack.is_empty() else None
                # If the current node has not been fulfilled yet, we don't need to check for loop
                if current_latest_fulfilled_id is not None:
                    # Trace back through the dependency chain to detect if this node triggers itself
                    visited = set()
                    current_execution_id = invoked_by

                    # Walk backwards through the dependency chain
                    while current_execution_id and current_execution_id not in visited:
                        visited.add(current_execution_id)

                        # Get the dependencies that triggered this execution
                        dependencies_for_current = state.meta.node_execution_cache._dependencies_invoked.get(
                            current_execution_id, set()
                        )

                        # If we've reached the end of the chain, it means the node is not in a loop
                        # we can break out of the loop
                        if not dependencies_for_current:
                            break

                        # Move to the previous node in the dependency chain
                        current_execution_id = next(iter(dependencies_for_current))

                        current_node_class = state.meta.node_execution_cache.__node_execution_lookup__.get(
                            current_execution_id
                        )

                        # If we've found our target node class in the chain
                        if current_node_class == cls.node_class:
                            # Check if the execution id is the same as the latest fulfilled execution id
                            # If yes, we're in a loop
                            if current_execution_id == current_latest_fulfilled_id:
                                in_loop = True
                            # If not, current node has been triggered by other node,
                            # we can break out of the loop
                            break

            for queued_node_execution_id in state.meta.node_execution_cache._node_executions_queued[cls.node_class]:
                if (
                    invoked_by not in state.meta.node_execution_cache._dependencies_invoked[queued_node_execution_id]
                    and not in_loop
                ):
                    state.meta.node_execution_cache._invoke_dependency(
                        queued_node_execution_id, cls.node_class, invoked_by, dependencies
                    )
                    return queued_node_execution_id

            state.meta.node_execution_cache._node_executions_queued[cls.node_class].append(execution_id)
            state.meta.node_execution_cache._invoke_dependency(execution_id, cls.node_class, invoked_by, dependencies)
            return execution_id

    class Execution(metaclass=_BaseNodeExecutionMeta):
        node_class: Type["BaseNode"]
        count: int

    def __init__(
        self,
        *,
        state: Optional[StateType] = None,
        context: Optional[WorkflowContext] = None,
    ):
        if state:
            self.state = state
        else:
            # Instantiate a default state class if one is not provided, for ease of testing

            original_base = get_original_base(self.__class__)

            args = get_args(original_base)

            if args and len(args) > 0:
                state_type = args[0]
                if isinstance(state_type, TypeVar):
                    state_type = BaseState
            else:
                state_type = BaseState

            self.state = state_type()

        self._context = context or WorkflowContext()
        inputs: Dict[str, Any] = {}
        for descriptor in self.__class__:
            if not descriptor.instance:
                continue

            if any(isinstance(t, type) and issubclass(t, BaseDescriptor) for t in descriptor.types):
                # We don't want to resolve attributes that are _meant_ to be descriptors
                continue

            resolved_value = resolve_value(descriptor.instance, self.state, path=descriptor.name, memo=inputs)
            setattr(self, descriptor.name, resolved_value)

        # We only want to store the attributes that were actually set as inputs, not every attribute that exists.
        all_inputs = {}
        for key, value in inputs.items():
            path_parts = key.split(".")
            node_attribute_descriptor = getattr(self.__class__, path_parts[0])
            inputs_key = reduce(lambda acc, part: acc[part], path_parts[1:], node_attribute_descriptor)
            all_inputs[inputs_key] = value

        self._inputs = MappingProxyType(all_inputs)

    def run(self) -> NodeRunResponse:
        return self.Outputs()

    def __repr__(self) -> str:
        return str(self.__class__)

    __simulates_workflow_output__ = False

    def __directly_emit_workflow_output__(
        self, event: NodeExecutionStreamingEvent, workflow_output_descriptor: OutputReference
    ) -> bool:
        """
        In the legacy workflow runner, there was support for emitting streaming workflow outputs for prompt nodes
        connected to terminal nodes. These two private methods provides a hacky, intentionally short-lived workaround
        for us to enable this until we can directly reference prompt outputs from the UI.
        """

        return False



# This file was auto-generated by Fern from our API Definition.

import typing
from .terminal_node_string_result import TerminalNodeStringResult
from .terminal_node_number_result import TerminalNodeNumberResult
from .terminal_node_json_result import TerminalNodeJsonResult
from .terminal_node_chat_history_result import TerminalNodeChatHistoryResult
from .terminal_node_search_results_result import TerminalNodeSearchResultsResult
from .terminal_node_error_result import TerminalNodeErrorResult
from .terminal_node_array_result import TerminalNodeArrayResult
from .terminal_node_function_call_result import TerminalNodeFunctionCallResult

TerminalNodeResultOutput = typing.Union[
    TerminalNodeStringResult,
    TerminalNodeNumberResult,
    TerminalNodeJsonResult,
    TerminalNodeChatHistoryResult,
    TerminalNodeSearchResultsResult,
    TerminalNodeErrorResult,
    TerminalNodeArrayResult,
    TerminalNodeFunctionCallResult,
]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value import ArrayVellumValue
from .vellum_value import VellumValue
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import typing
import pydantic


class CodeExecutorResponse(UniversalBaseModel):
    log: str
    output: VellumValue

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import asyncio
import json
import logging
from typing import Any, Dict, List, Optional

import httpx

from vellum.workflows.constants import AuthorizationType
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.types.core import VellumSecret
from vellum.workflows.types.definition import MCPServer, MCPToolDefinition

logger = logging.getLogger(__name__)


class MCPHttpClient:
    """
    Direct HTTP implementation for MCP (Model Context Protocol) client
    without using the official Python SDK.

    Supports Streamable HTTP transport using httpx.
    """

    def __init__(self, server_url: str, headers: Dict[str, str], session_timeout: int = 30):
        """
        Initialize MCP HTTP client.

        Args:
            server_url: The MCP server endpoint URL (e.g., "https://example.com/mcp")
            headers: Authentication headers
            session_timeout: Timeout for HTTP requests in seconds
        """
        self.server_url = server_url.rstrip("/")
        self.headers = headers
        self.session_timeout = session_timeout
        self.session_id: Optional[str] = None
        self.request_id = 0
        self._client: Optional[httpx.AsyncClient] = None

    async def __aenter__(self):
        """Async context manager entry."""
        self._client = httpx.AsyncClient(timeout=httpx.Timeout(self.session_timeout), headers=self.headers)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        if self._client:
            await self._client.aclose()

    def _next_request_id(self) -> int:
        """Generate next request ID."""
        self.request_id += 1
        return self.request_id

    async def _send_request(self, method: str, params: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Send a JSON-RPC request to the MCP server.

        Args:
            method: The JSON-RPC method name
            params: Optional parameters for the method

        Returns:
            The JSON-RPC response
        """
        if not self._client:
            raise RuntimeError("Client session not initialized. Use 'async with' context manager.")

        # Prepare JSON-RPC request
        request_data = {"jsonrpc": "2.0", "id": self._next_request_id(), "method": method, "params": params or {}}

        # Prepare headers
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        # Include session ID if we have one
        if self.session_id:
            headers["Mcp-Session-Id"] = self.session_id

        logger.debug(f"Sending request: {json.dumps(request_data, indent=2)}")

        # Send POST request
        response = await self._client.post(self.server_url, json=request_data, headers=headers)

        # Check for session ID in response headers
        if "Mcp-Session-Id" in response.headers:
            self.session_id = response.headers["Mcp-Session-Id"]
            logger.debug(f"Received session ID: {self.session_id}")

        # Handle JSON response
        response_data = response.json()
        logger.debug(f"Received response: {json.dumps(response_data, indent=2)}")

        if "error" in response_data:
            raise Exception(f"MCP Error: {response_data['error']}")

        return response_data

    async def initialize(self, client_info: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Initialize the MCP connection.

        Args:
            client_info: Optional client information

        Returns:
            Server capabilities and information
        """
        params = {
            "protocolVersion": "2025-06-18",
            "capabilities": {},
            "clientInfo": client_info or {"name": "vellum-mcp-client", "version": "1.0.0"},
        }

        response = await self._send_request("initialize", params)
        return response.get("result", {})

    async def list_tools(self) -> List[Dict[str, Any]]:
        """
        Get list of available tools from the server.

        Returns:
            List of tool definitions
        """
        response = await self._send_request("tools/list")
        return response.get("result", {}).get("tools", [])

    async def call_tool(self, name: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """
        Call a tool on the server.

        Args:
            name: Tool name
            arguments: Tool arguments

        Returns:
            Tool execution result
        """
        params = {"name": name, "arguments": arguments}

        response = await self._send_request("tools/call", params)
        return response.get("result", {})


class MCPService:
    def _get_auth_headers(self, server: MCPServer) -> Dict[str, str]:
        headers = {}
        if server.authorization_type == AuthorizationType.BEARER_TOKEN:
            token = server.bearer_token_value
            if not token:
                raise NodeException(
                    "Bearer token is required for BEARER_TOKEN authorization type",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                )

            headers["Authorization"] = f"Bearer {token}"

        elif server.authorization_type == AuthorizationType.API_KEY:
            key = server.api_key_header_key
            value = server.api_key_header_value
            if not key or not value:
                raise NodeException(
                    "API key header key and value are required for API_KEY authorization type",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                )
            if isinstance(value, VellumSecret):
                headers[key] = value.name
            elif isinstance(value, str):
                headers[key] = value

        return headers

    async def _execute_mcp_call(self, server: MCPServer, operation: str, **kwargs) -> Any:
        """Execute an MCP operation using direct HTTP calls."""
        headers = self._get_auth_headers(server)

        try:
            async with MCPHttpClient(server.url, headers) as client:
                await client.initialize()

                if operation == "list_tools":
                    return await client.list_tools()
                elif operation == "call_tool":
                    return await client.call_tool(
                        name=kwargs["name"],
                        arguments=kwargs["arguments"],
                    )
                else:
                    raise ValueError(f"Unknown MCP operation: {operation}")

        except Exception as e:
            logger.error(f"Error executing MCP operation {operation}: {e}")
            raise NodeException(
                message=f"Error executing MCP operation '{operation}': {str(e)}",
                code=WorkflowErrorCode.NODE_EXECUTION,
            )

    def list_tools(self, server: MCPServer) -> List[Dict[str, Any]]:
        """List available tools from an MCP server."""
        try:
            tools = asyncio.run(self._execute_mcp_call(server, "list_tools"))
            return tools
        except Exception as e:
            logger.warning(f"Failed to list tools from MCP server '{server.name}': {e}")
            return []

    def execute_tool(self, tool_def: MCPToolDefinition, arguments: Dict[str, Any]) -> Any:
        """Execute a tool on an MCP server."""
        try:
            result = asyncio.run(
                self._execute_mcp_call(
                    tool_def.server,
                    "call_tool",
                    name=tool_def.name,
                    arguments=arguments,
                )
            )
            return result
        except Exception as e:
            logger.error(f"Error executing MCP tool '{tool_def.name}': {e}")
            raise NodeException(
                message=f"Error executing MCP tool '{tool_def.name}': {str(e)}",
                code=WorkflowErrorCode.NODE_EXECUTION,
            )

    def hydrate_tool_definitions(self, server_def: MCPServer) -> List[MCPToolDefinition]:
        """Hydrate an MCPToolDefinition with detailed information from the MCP server."""
        try:
            tools = self.list_tools(server_def)

            return [
                MCPToolDefinition(
                    name=tool["name"],
                    server=server_def,
                    description=tool["description"],
                    parameters=tool["inputSchema"],
                )
                for tool in tools
            ]
        except Exception as e:
            logger.warning(f"Failed to hydrate MCP server '{server_def.name}': {e}")
            return []



from uuid import UUID
from typing import Any, List, Optional, Type, Union, cast

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.references import NodeReference
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.exceptions import UnsupportedSerializationException
from vellum_ee.workflows.display.utils.expressions import get_child_descriptor
from vellum_ee.workflows.display.utils.vellum import (
    ConstantValuePointer,
    ExecutionCounterData,
    ExecutionCounterPointer,
    InputVariableData,
    InputVariablePointer,
    NodeInputValuePointerRule,
    WorkspaceSecretData,
    WorkspaceSecretPointer,
    create_node_input_value_pointer_rule,
)
from vellum_ee.workflows.display.vellum import NodeInput, NodeInputValuePointer


def create_node_input(
    node_id: UUID,
    input_name: str,
    value: Any,
    display_context: WorkflowDisplayContext,
    input_id: Union[Optional[UUID], Optional[str]] = None,
    pointer_type: Optional[Type[NodeInputValuePointerRule]] = ConstantValuePointer,
) -> NodeInput:
    input_id = str(input_id) if input_id else str(uuid4_from_hash(f"{node_id}|{input_name}"))
    rules = create_node_input_value_pointer_rules(value, display_context, pointer_type=pointer_type)
    return NodeInput(
        id=input_id,
        key=input_name,
        value=NodeInputValuePointer(
            rules=rules,
            combinator="OR",
        ),
    )


def create_node_input_value_pointer_rules(
    value: Any,
    display_context: WorkflowDisplayContext,
    existing_rules: Optional[List[NodeInputValuePointerRule]] = None,
    pointer_type: Optional[Type[NodeInputValuePointerRule]] = None,
) -> List[NodeInputValuePointerRule]:
    node_input_value_pointer_rules: List[NodeInputValuePointerRule] = existing_rules or []

    if isinstance(value, BaseDescriptor):
        if isinstance(value, NodeReference):
            if value.instance is None:
                raise ValueError(f"Expected NodeReference {value.name} to have an instance")
            value = cast(BaseDescriptor, value.instance)

        if isinstance(value, LazyReference):
            child_descriptor = get_child_descriptor(value, display_context)
            return create_node_input_value_pointer_rules(
                child_descriptor, display_context, [], pointer_type=pointer_type
            )

        if isinstance(value, CoalesceExpression):
            # Recursively handle the left-hand side
            lhs_rules = create_node_input_value_pointer_rules(value.lhs, display_context, [], pointer_type=pointer_type)
            node_input_value_pointer_rules.extend(lhs_rules)

            # Handle the right-hand side
            if not isinstance(value.rhs, CoalesceExpression):
                rhs_rules = create_node_input_value_pointer_rules(
                    value.rhs, display_context, [], pointer_type=pointer_type
                )
                node_input_value_pointer_rules.extend(rhs_rules)
        else:
            # Non-CoalesceExpression case
            try:
                rule = create_node_input_value_pointer_rule(value, display_context)
            except UnsupportedSerializationException:
                return node_input_value_pointer_rules

            node_input_value_pointer_rules.append(rule)
    else:
        pointer = create_pointer(value, pointer_type)
        node_input_value_pointer_rules.append(pointer)

    return node_input_value_pointer_rules


def create_pointer(
    value: Any,
    pointer_type: Optional[Type[NodeInputValuePointerRule]] = None,
) -> NodeInputValuePointerRule:
    if value is None:
        if pointer_type is WorkspaceSecretPointer:
            return WorkspaceSecretPointer(
                type="WORKSPACE_SECRET", data=WorkspaceSecretData(type="STRING", workspace_secret_id=None)
            )

    vellum_variable_value = primitive_to_vellum_value(value)
    if pointer_type is InputVariablePointer:
        return InputVariablePointer(type="INPUT_VARIABLE", data=InputVariableData(input_variable_id=value))
    elif pointer_type is WorkspaceSecretPointer:
        return WorkspaceSecretPointer(
            type="WORKSPACE_SECRET", data=WorkspaceSecretData(type="STRING", workspace_secret_id=value)
        )
    elif pointer_type is ExecutionCounterPointer:
        return ExecutionCounterPointer(type="EXECUTION_COUNTER", data=ExecutionCounterData(node_id=value))
    elif pointer_type is ConstantValuePointer or pointer_type is None:
        return ConstantValuePointer(type="CONSTANT_VALUE", data=vellum_variable_value)
    else:
        raise ValueError(f"Pointer type {pointer_type} not supported")



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from .chat_message_request import ChatMessageRequest
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class WorkflowExecutionActualChatHistoryRequest(UniversalBaseModel):
    output_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The Vellum-generated ID of a workflow output. Must provide either this or output_key. output_key is typically preferred.
    """

    output_key: typing.Optional[str] = pydantic.Field(default=None)
    """
    The user-defined name of a workflow output. Must provide either this or output_id. Should correspond to the `Name` specified in a Final Output Node. Generally preferred over output_id.
    """

    quality: typing.Optional[float] = pydantic.Field(default=None)
    """
    Optionally provide a decimal number between 0.0 and 1.0 (inclusive) representing the quality of the output. 0 is the worst, 1 is the best.
    """

    metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    Optionally provide additional metadata about the feedback submission.
    """

    timestamp: typing.Optional[float] = pydantic.Field(default=None)
    """
    Optionally provide the timestamp representing when this feedback was collected. Used for reporting purposes.
    """

    output_type: typing.Literal["CHAT_HISTORY"] = "CHAT_HISTORY"
    desired_output_value: typing.Optional[typing.List[ChatMessageRequest]] = pydantic.Field(default=None)
    """
    Optionally provide the value that the output ideally should have been.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .test_suite_test_case_create_bulk_operation_request import TestSuiteTestCaseCreateBulkOperationRequest
from .test_suite_test_case_replace_bulk_operation_request import TestSuiteTestCaseReplaceBulkOperationRequest
from .test_suite_test_case_upsert_bulk_operation_request import TestSuiteTestCaseUpsertBulkOperationRequest
from .test_suite_test_case_delete_bulk_operation_request import TestSuiteTestCaseDeleteBulkOperationRequest

TestSuiteTestCaseBulkOperationRequest = typing.Union[
    TestSuiteTestCaseCreateBulkOperationRequest,
    TestSuiteTestCaseReplaceBulkOperationRequest,
    TestSuiteTestCaseUpsertBulkOperationRequest,
    TestSuiteTestCaseDeleteBulkOperationRequest,
]



from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from vellum.client.core.api_error import ApiError
from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.array_vellum_value import ArrayVellumValue
from vellum.client.types.vellum_value import VellumValue
from vellum.client.types.vellum_variable_type import VellumVariableType
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.references import OutputReference, WorkflowInputReference
from vellum.workflows.references.execution_count import ExecutionCountReference
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.references.node import NodeReference
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum.workflows.utils.vellum_variables import primitive_type_to_vellum_variable_type
from vellum_ee.workflows.display.utils.exceptions import UnsupportedSerializationException
from vellum_ee.workflows.display.utils.expressions import get_child_descriptor

if TYPE_CHECKING:
    from vellum_ee.workflows.display.types import WorkflowDisplayContext


class ConstantValuePointer(UniversalBaseModel):
    type: Literal["CONSTANT_VALUE"] = "CONSTANT_VALUE"
    data: VellumValue


ArrayVellumValue.model_rebuild()


class NodeOutputData(UniversalBaseModel):
    node_id: str
    output_id: str


class NodeOutputPointer(UniversalBaseModel):
    type: Literal["NODE_OUTPUT"] = "NODE_OUTPUT"
    data: NodeOutputData


class InputVariableData(UniversalBaseModel):
    input_variable_id: str


class InputVariablePointer(UniversalBaseModel):
    type: Literal["INPUT_VARIABLE"] = "INPUT_VARIABLE"
    data: InputVariableData


class WorkspaceSecretData(UniversalBaseModel):
    type: VellumVariableType
    workspace_secret_id: Optional[str] = None


class WorkspaceSecretPointer(UniversalBaseModel):
    type: Literal["WORKSPACE_SECRET"] = "WORKSPACE_SECRET"
    data: WorkspaceSecretData


class ExecutionCounterData(UniversalBaseModel):
    node_id: str


class ExecutionCounterPointer(UniversalBaseModel):
    type: Literal["EXECUTION_COUNTER"] = "EXECUTION_COUNTER"
    data: ExecutionCounterData


NodeInputValuePointerRule = Union[
    NodeOutputPointer,
    InputVariablePointer,
    ConstantValuePointer,
    WorkspaceSecretPointer,
    ExecutionCounterPointer,
]


def infer_vellum_variable_type(value: Any) -> VellumVariableType:
    inferred_type: VellumVariableType

    if isinstance(value, BaseDescriptor):
        descriptor: BaseDescriptor = value
        if isinstance(descriptor, NodeReference):
            if not isinstance(descriptor.instance, BaseDescriptor):
                raise ValueError(
                    f"Expected NodeReference {descriptor.name} to have an instance pointing to a descriptor"
                )
            descriptor = descriptor.instance

        inferred_type = primitive_type_to_vellum_variable_type(descriptor)
    else:
        vellum_variable_value = primitive_to_vellum_value(value)
        inferred_type = vellum_variable_value.type

    return inferred_type


def create_node_input_value_pointer_rule(
    value: Any, display_context: "WorkflowDisplayContext"
) -> NodeInputValuePointerRule:
    if isinstance(value, OutputReference):
        if value not in display_context.global_node_output_displays:
            if issubclass(value.outputs_class, BaseNode.Outputs):
                raise ValueError(f"Reference to node '{value.outputs_class._node_class.__name__}' not found in graph.")

            raise ValueError(f"Reference to outputs '{value.outputs_class.__qualname__}' is invalid.")

        upstream_node, output_display = display_context.global_node_output_displays[value]
        upstream_node_display = display_context.global_node_displays[upstream_node]
        return NodeOutputPointer(
            data=NodeOutputData(node_id=str(upstream_node_display.node_id), output_id=str(output_display.id)),
        )
    if isinstance(value, LazyReference):
        child_descriptor = get_child_descriptor(value, display_context)
        return create_node_input_value_pointer_rule(child_descriptor, display_context)
    if isinstance(value, WorkflowInputReference):
        workflow_input_display = display_context.global_workflow_input_displays[value]
        return InputVariablePointer(data=InputVariableData(input_variable_id=str(workflow_input_display.id)))
    if isinstance(value, VellumSecretReference):
        try:
            workspace_secret = display_context.client.workspace_secrets.retrieve(
                id=value.name,
            )
            workspace_secret_id: Optional[str] = str(workspace_secret.id)
        except ApiError:
            workspace_secret_id = None

        return WorkspaceSecretPointer(
            data=WorkspaceSecretData(
                type="STRING",
                workspace_secret_id=workspace_secret_id,
            ),
        )
    if isinstance(value, ExecutionCountReference):
        node_class_display = display_context.node_displays[value.node_class]
        return ExecutionCounterPointer(
            data=ExecutionCounterData(node_id=str(node_class_display.node_id)),
        )

    if not isinstance(value, BaseDescriptor):
        vellum_value = primitive_to_vellum_value(value)
        return ConstantValuePointer(type="CONSTANT_VALUE", data=vellum_value)

    raise UnsupportedSerializationException(f"Unsupported descriptor type: {value.__class__.__name__}")



# This file was auto-generated by Fern from our API Definition.

import typing
from .node_execution_initiated_event import NodeExecutionInitiatedEvent
from .node_execution_streaming_event import NodeExecutionStreamingEvent
from .node_execution_fulfilled_event import NodeExecutionFulfilledEvent
from .node_execution_rejected_event import NodeExecutionRejectedEvent
from .node_execution_paused_event import NodeExecutionPausedEvent
from .node_execution_resumed_event import NodeExecutionResumedEvent

VellumNodeExecutionEvent = typing.Union[
    NodeExecutionInitiatedEvent,
    NodeExecutionStreamingEvent,
    NodeExecutionFulfilledEvent,
    NodeExecutionRejectedEvent,
    NodeExecutionPausedEvent,
    NodeExecutionResumedEvent,
]



# This file was auto-generated by Fern from our API Definition.

from ...core.client_wrapper import SyncClientWrapper
import typing
from ...core.request_options import RequestOptions
from ...core.http_response import HttpResponse
from ...types.organization_read import OrganizationRead
from ...core.pydantic_utilities import parse_obj_as
from json.decoder import JSONDecodeError
from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper
from ...core.http_response import AsyncHttpResponse


class RawOrganizationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def organization_identity(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[OrganizationRead]:
        """
        Retrieves information about the active Organization

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OrganizationRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/organizations/identity",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationRead,
                    parse_obj_as(
                        type_=OrganizationRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncRawOrganizationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def organization_identity(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[OrganizationRead]:
        """
        Retrieves information about the active Organization

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OrganizationRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/organizations/identity",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationRead,
                    parse_obj_as(
                        type_=OrganizationRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)
